'''
MODIS sinusoidal projection forward and backward coordinate transformations,
courtesy of Giglio et al. (2018), Collection 6 MODIS Burned Area Product
User's Guide, Version 1, Appendix B:

    https://modis-fire.umd.edu/files/MODIS_C6_BA_User_Guide_1.2.pdf
'''

import re
import numpy as np

SPHERE_RADIUS = 6371007.181 # Radius of ideal sphere, meters
TILE_LINES = {250: 5000, 500: 2400, 1000: 1200} # Num. lines by nominal res.
TILE_SIZE = 1111950 # Width and height of MODIS tile in projection plane
XMIN = -20015109 # Western limit of projection plane, meters
YMAX = 10007555 # Northern limit of projection plane, meters
VIIRS_METADATA = re.compile(
    r'.*XDim=(?P<xdim>\d+)'
    r'.*YDim=(?P<ydim>\d+)'
    r'.*UpperLeftPointMtrs=\((?P<ul>[0-9,\-\.]+)\)'
    r'.*LowerRightMtrs=\((?P<lr>[0-9,\-\.]+)\).*', re.DOTALL)
VIIRS_H5_ROOT = 'HDFEOS/GRIDS/VNP_Grid_VNP15A2H/Data Fields'

def dec2bin_unpack(x):
    '''
    Unpacks an arbitrary decimal NumPy array into a binary representation
    along a new axis. Assumes decimal digits are on the interval [0, 255],
    i.e., that only 8-bit representations are needed.

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    '''
    # Make sure the bit representation is enumerated along a new axis, the
    #   very last axis
    axis = x.ndim
    # unpackbits() returns the bit representation in big-endian order, so we
    #   flip the array (with -8) to get litte-endian order
    return np.unpackbits(x[...,None], axis = axis)[...,-8:]


def geotransform(
        hdf, ps = 463.31271653, nrow = 2400, ncol = 2400,
        metadata = VIIRS_METADATA):
    '''
    Prescribe a geotransform tuple for the output GeoTIFF. For MODIS/VIIRS
    sinsuoidal projections, the lower right corner coordinates are "the only
    metadata that accurately reflect the extreme corners of the gridded image"
    (Myneni et al. 2018, VIIRS LAI/fPAR User Guide). So, instead of using the
    reported upper-left (UL) corner coordinates, it is more accurate to use
    the lower-right (LR) corner coordinates and calculate the position of the
    UL corner based on the width and height of the image and the pixel size.
    NOTE that a rather odd pixel size is required to get the correct results
    verified against the HDF-EOS-to-GeoTIFF (HEG) Conversion Tool; see also
    Giglio et al. (2018), "Collection 6 MODIS Burned Area Product User's
    Guide, Version 1" Table 1.

        https://modis-land.gsfc.nasa.gov/pdf/MODIS_C6_BA_User_Guide_1.2.pdf

    Parameters
    ----------
    hdf : h5py.File
    ps : int or float
        The pixel size; in units matching the linear units of the SRS
        (Default: 463.3127 meters)
    nrow : int
        Number of rows in the output image (Default: 2400 for MODIS/VIIRS)
    ncol : int
        Number of columns in the output image (Default: 2400 for MODIS/VIIRS)
    metadata : re.Pattern
        Compiled regex that captures important metadata fields

    Returns
    -------
    tuple
        (x_min, pixel_width, 0, y_max, 0, -pixel_height)
    '''
    meta = hdf['HDFEOS INFORMATION/StructMetadata.0'][()].decode()
    lr = VIIRS_METADATA.match(meta).groupdict()['lr'].split(',')
    return ( # Subtract distance (meters) from LR corner to obtain UR corner
        float(lr[0]) - (ncol * ps), ps, 0, float(lr[1]) + (nrow * ps), 0, -ps)


def mod15a2h_qc_fail(x):
    '''
    Returns pass/fail for QC flags based on the L4C fPAR QC protocol for the
    `FparLai_QC` band: Bad pixels have either `1` in the first bit ("Pixel not
    produced at all") or anything other than `00` ("clear") in bits 3-4.
    Output array is True wherever the array fails QC criteria. Compare to:

        np.vectorize(lambda v: v[0] == 1 or v[3:5] != '00')

    Parameters
    ----------
    x : numpy.ndarray
        Array where the last axis enumerates the unpacked bits
        (ones and zeros)

    Returns
    -------
    numpy.ndarray
        Boolean array with True wherever QC criteria are failed
    '''
    y = dec2bin_unpack(x)
    # Emit 1 = FAIL if these two bits are not == "00"
    c1 = y[...,3:5].sum(axis = -1).astype(np.uint8)
    # Emit 1 = FAIL if 1st bit == 1 ("Pixel not produced at all")
    c2 = y[...,0]
    # Intermediate arrays are 1 = FAIL, 0 = PASS
    return (c1 + c2) > 0


def modis_from_wgs84(coords):
    '''
    Given longitude-latitude coordinates, return the coordinates on the
    sinusoidal projection plane.

    Parameters
    ----------
    coords : tuple or list
        (Longitude, Latitude) coordinate pair

    Returns
    -------
    tuple
        (X, Y) coordinate pair in MODIS sinusoidal projection
    '''
    x, y = map(np.deg2rad, coords)
    return (SPHERE_RADIUS * x * np.cos(y), SPHERE_RADIUS * y)


def modis_to_wgs84(coords):
    '''
    Convert coordinates on the MODIS sinusoidal plane to longitude-latitude
    coordinates (WGS84).

    Parameters
    ----------
    coords : tuple or list
        (X, Y) coordinate pair in MODIS sinusoidal projection

    Returns
    -------
    tuple
        (Longitude, Latitude) coordinate pair
    '''
    x, y = coords
    lat = y / SPHERE_RADIUS
    lng = x / (SPHERE_RADIUS * np.cos(lat))
    return tuple(map(np.rad2deg, (lng, lat)))


def modis_tile_from_wgs84(coords):
    '''
    Given longitude-latitude coordinates, return the MODIS tile (H,V) that
    contains them.

    Parameters
    ----------
    coords : tuple or list
        (Longitude, Latitude) coordinate pair

    Returns
    -------
    tuple
        (H,V) tile identifier
    '''
    x, y = modis_from_wgs84(coords) # Get coordinates in the projection plane
    return (
        np.floor((x - XMIN) / TILE_SIZE),
        np.floor((YMAX - y) / TILE_SIZE))


def modis_row_col_from_wgs84(coords, nominal = 500):
    '''
    Given longitude-latitude coordinates, return the corresponding row-column
    coordinates within a MODIS tile. NOTE: You'll need to determine which
    MODIS tile contains this row-column index with `modis_tile_from_wgs84()`.

    Parameters
    ----------
    coords : tuple or list
        (X, Y) coordinate pair in MODIS sinusoidal projection
    nominal : int
        Nominal resolution of MODIS raster: 250 (meters), 500, or 1000

    Returns
    -------
    tuple
        (Row, Column) coordinates
    '''
    x, y = modis_from_wgs84(coords) # Get coordinates in the projection plane
    # Get actual size of, e.g., "500-m" MODIS sinusoidal cell
    res = TILE_SIZE / float(TILE_LINES[nominal])
    return (
        np.floor((((YMAX - y) % TILE_SIZE) / res) - 0.5),
        np.floor((((x - XMIN) % TILE_SIZE) / res) - 0.5),
    )


def modis_row_col_to_wgs84(coords, h, v, nominal = 500):
    '''
    Convert pixel coordinates in a specific MODIS tile to longitude-latitude
    coordinates.

    Parameters
    ----------
    coords : tuple or list
        (X, Y) coordinate pair in MODIS sinusoidal projection
    h : int
        MODIS tile "h" index
    v : int
        MODIS tile "v" index
    nominal : int
        Nominal resolution of MODIS raster: 250 (meters), 500, or 1000

    Returns
    -------
    tuple
        (Longitude, Latitude) coordinates
    '''
    r, c = coords
    lines = TILE_LINES[nominal]
    assert (0 <= r <= (lines - 1)) and (0 <= c <= (lines - 1)),\
        'Row and column indices must be in the range [0, %d]' % (lines - 1)
    # Get actual size of, e.g., "500-m" MODIS sinusoidal cell
    res = TILE_SIZE / float(TILE_LINES[nominal])
    x = ((c + 0.5) * res) + (h * TILE_SIZE) + XMIN
    y = YMAX - ((r + 0.5) * res) - (v * TILE_SIZE)
    return modis_to_wgs84((x, y))


def vnp15a2h_qc_fail(x):
    '''
    Returns pass/fail for QC flags based on the L4C fPAR QC protocol for the
    `FparLai_QC` band: Bad pixels have either `11` in the first two bits
    ("Fill Value") or anything other than `0` in the 3rd least-significant
    bits, which combines "Pixel not produced at all". For example, see decimal
    number 80:

        0101|0|000 where "000" is the combined (Fill bit | Retrieval quality)

    Parameters
    ----------
    x : numpy.ndarray
        Unsigned, 8-bit integer array

    Returns
    -------
    numpy.ndarray
        Boolean array
    '''
    y = dec2bin_unpack(x)
    # Emit 1 = FAIL if sum("11") == 2; "BiomeType" == "Filled Value"
    c1 = np.where(y[...,0:2].sum(axis = -1) == 2, 1, 0).astype(np.uint8)
    # Emit 1 = FAIL if 3rd bit == 1; "SCF_QC" == "Pixel not produced at all"
    c2 = y[...,5]
    # Intermediate arrays are 1 = FAIL, 0 = PASS
    return (c1 + c2) > 0


def vnp15a2h_cloud_fail(x):
    '''
    Returns pass/fail for QC flags based on the L4C fPAR QC protocol for the
    `FparExtra_QC` band (cloud QC band): Bad pixels have anything OTHER THAN
    `1` second least-significant bit; `00` and `01` being acceptable cloud QC
    flags ("Confident clear" and "Probably clear", respectively).

    Parameters
    ----------
    x : numpy.ndarray
        Unsigned, 8-bit integer array

    Returns
    -------
    numpy.ndarray
        Boolean array
    '''
    y = dec2bin_unpack(x)
    return y[...,-2] > 0



'''
Commonly used global constants. **New users should change
`ANCILLARY_DATA_PATHS` if it doesn't match their file system.**

Source for EASE-Grid 2.0 parameters:

    https://nsidc.org/ease/ease-grid-projection-gt

NOTE: EASE-Grid 2.0 parameters do not match exactly those specified at NSIDC
because, in fact, they must be 1000 or 9000, depending on the grid size, in
order to get the right number of rows and columns in the output image.
Grid resolution is only a whole number for the polar (north or south
hemisphere) grids.
'''

import csv
import numpy as np
from collections import OrderedDict

ANCILLARY_DATA_PATHS = {
    'smap_l4c_ancillary_data_file_path': 'SPL4C_Vv4040_SMAP_L4_C.Ancillary.h5',
    'smap_l4c_1km_ancillary_data_lc_path': 'MCD12Q1_M01_lc_dom_uint8',
    'smap_l4c_9km_ancillary_data_lc_path': 'MOD12Q1_M09_lc_dom_uint8',
    'smap_l4c_1km_ancillary_data_x_coord_path': 'SMAP_L4_C_LON_14616_x_34704_M01_flt32',
    'smap_l4c_1km_ancillary_data_y_coord_path': 'SMAP_L4_C_LAT_14616_x_34704_M01_flt32',
    'smap_l4c_9km_ancillary_data_x_coord_path': 'SMAP_L4_C_LON_1624_x_3856_M09_flt32',
    'smap_l4c_9km_ancillary_data_y_coord_path': 'SMAP_L4_C_LAT_1624_x_3856_M09_flt32',
    'smap_l4c_9km_pft_subgrid_counts_CONUS': 'SMAP_L4C_Vv4040_1km_subgrid_PFT_counts_CONUS.h5',
    'smap_l4c_9km_sparse_col_index': 'MCD12Q1_M09land_col.uint16',
    'smap_l4c_9km_sparse_row_index': 'MCD12Q1_M09land_row.uint16',
    'transcom_netcdf_path': 'CarbonTracker_TransCom_and_other_regions.nc'
}


# Version 4020 BPLUT parameters in order of PFT numeric code (PFT 0 through 9)
BPLUT = OrderedDict({
    # Only reason for OrderedDict (instead of dict) is for consistency with
    #   pyl4c.data.fixtures.BPLUT()
    '_version': '4',
    'LUE': np.array([[ # gC per MJ
        np.nan, 1.71, 1.38, 1.71, 1.19, 1.95, 1.68, 2.53, 3.6, np.nan
    ]]),
    'CUE': np.array([[
        np.nan, 0.7, 0.57, 0.79, 0.78, 0.65, 0.6, 0.72, 0.71, np.nan
    ]]),
    'tmin': np.array([ # degrees K
        [np.nan, 235, 230, 259, 260, 259, 251, 246, 266, np.nan],
        [np.nan, 309, 303, 301, 284, 304, 302, 314, 319, np.nan]
    ]),
    'vpd': np.array([ # Pascals
        [np.nan, 0, 15, 869, 1500, 4, 0, 228, 1500, np.nan],
        [np.nan, 4169, 7000, 3452, 5401, 4282, 4229, 4516, 7000, np.nan]
    ]),
    'smrz': np.array([ # Percent saturation
        [np.nan, -30, 18, -30, -26, 22, -30, -15, 10, np.nan],
        [np.nan, 74, 26, 49, 87, 72, 76, 30, 68, np.nan]
    ]),
    'smsf': np.array([ # Percent saturation
        [np.nan, -50, -46, -9, -50, -15, -3, -42, -49, np.nan],
        [np.nan, 46, 59, 43, 53, 48, 51, 41, 30, np.nan]
    ]),
    'ft': np.array([ # Frozen = 0, Thawed = 1
        [np.nan, 0.58, 0.36, 0.55, 0.9, 0.91, 0.85, 0.77, 1, np.nan],
        [np.nan, 1, 1, 1, 1, 1, 1, 1, 1, np.nan]
    ]),
    'tsoil': np.array([[ # degrees K
        np.nan, 265.04, 477.83, 238.81, 267.26, 292.82, 232.24, 263.97, 329.63, np.nan
    ]]),
    # NOTE: The medium/structural and slow/recalcitrant decay constants
    #   (2nd and 3rd rows) are true decay constants, unlike in Ops
    #   BPLUT which presents only dimensionless scalars.
    'decay_rates': np.array([ # days^-1
        [np.nan, 0.027, 0.028, 0.028, 0.03, 0.015, 0.025, 0.019, 0.035, np.nan],
        [np.nan, 1.08e-2, 1.12e-2, 1.12e-2, 1.2e-2, 0.6e-2, 1e-2, 0.76e-2, 1.4e-2, np.nan],
        [np.nan, 2.51e-4, 2.6e-4, 2.6e-4, 2.79e-4, 1.4e-4, 2.33e-4, 1.77e-4, 3.26e-4, np.nan]
    ]),
    'f_metabolic': np.array([[ # Fraction of daily litterfall entering "fast" or metabolic pool
        np.nan, 0.49, 0.71, 0.67, 0.67, 0.62, 0.76, 0.78, 0.78, np.nan
    ]]), # Also known as "f_met"
    'f_structural': np.array([[ # Fraction of structural (str) pool transferred in "humification"
        np.nan, 0.3, 0.3, 0.7, 0.3, 0.35, 0.55, 0.5, 0.8, np.nan
    ]]) # Also known as "f_str" -- See Jones et al. (2017, IEEE TGARS, p.5)
})


EASE2_GRID_PARAMS = {
    # A GeoTransform for a north-up raster is:
    #   (x_min, pixel_width, 0, y_max, 0, -pixel_height)
    'M01': {
        'epsg': 6933,
        'geotransform': (-17367530.45, 1000, 0, 7314540.83, 0, -1000),
        'resolution': 1000.89502334956, # From Brodzik et al.
        'shape': (14616, 34704),
        'size': 14616*34704
    },
    'M09': {
        'epsg': 6933,
        'geotransform': (-17367530.45, 9000, 0, 7314540.83, 0, -9000),
        'resolution': 9008.055210146, # From Brodzik et al.
        'shape': (1624, 3856),
        'size': 1624*3856
    },
    'N09': {
        'epsg': 6931,
        'geotransform': (-9000000.0, 9000, 0, 9000000.0, 0, -9000),
        'resolution': 9000,
        'shape': (2000, 2000),
        'size': 2000*2000
    },
    'M25': {
        'epsg': 6933,
        'geotransform': (-17367530.45, 25000, 0, 7307375.92, 0, -25000),
        'resolution': 25000,
        'shape': (584, 1388),
        'size': 584*1388
    },
    'M36': {
        'epsg': 6933,
        'geotransform': (-17367530.45, 36000, 0, 7314540.83, 0, -36000),
        'resolution': 36032.22,
        'shape': (406, 964),
        'size': 406*964
    }
}


HDF_PATHS = { # Where in the HDF hierarchy certain variables live
    'SPL4CMDL': { # By Earthdata Dataset ID
        '4': { # By Version number
            'longitude': 'GEO/longitude',
            'latitude': 'GEO/latitude',
            'SOC': 'SOC/soc_mean',
            'SOC*': 'SOC/soc_pft%d_mean',
            'GPP': 'GPP/gpp_mean',
            'GPP*': 'GPP/gpp_pft%d_mean',
            'NEE': 'NEE/nee_mean',
            'NEE*': 'NEE/nee_pft%d_mean',
            'RH': 'RH/rh_mean',
            'RH*': 'RH/rh_pft%d_mean',
        }
    },
    'SPL4SMGP': {
        '4': {'longitude': 'cell_lon', 'latitude': 'cell_lat',}
    }
}


PFT = OrderedDict({
    0: ('Water', 'WET'),
    1: ('Evergreen Needleleaf', 'ENF'),
    2: ('Evergreen Broadleaf', 'EBF'),
    3: ('Deciduous Needleleaf', 'DNF'),
    4: ('Deciduous Broadleaf', 'DBF'),
    5: ('Shrub', 'SHB'),
    6: ('Grass', 'GRS'),
    7: ('Cereal Crop', 'CCR'),
    8: ('Broadleaf Crop', 'BCR'),
    9: ('Urban and Built-Up', 'URB')
})


# BBOX extents, in decimal degrees, based on gdal_rasterize -te option:
#   https://gdal.org/programs/gdal_rasterize.html#cmdoption-gdal-translate-te
SUBSETS_BBOX = { # <xmin>, <ymin>, <xmax>, <ymax>
    'CONUS': [-124.5, 24.4, -66.7, 50.0],
    'WesternHemisphere': [-180, -90, 0, 90],
    'WesternHemisphere2': [-180, -90, -25, 90],
    'NorthernHemisphere': [-180, 0, 180, 90],
    'NorthernHemisphere45': [-180, 45, 180, 90],
    'NorthernHemisphere40': [-180, 40, 180, 90],
    'Nigeria': [2.3, 4, 14.8, 14],
    # AOI for: Chiodi and Harrison (2013) Journal of Climate 26 (3):822–837
    'ChiodiHarrison2013': [-160, -5, -110, 5],
    'Kang2014EMS': [-125, 47, -113, 49],
    'NorthernPlains': [-116.1, 40.9, -96.4, 49],
    'Montana': [-116.1, 44.4, -104.01, 49],
    'Iowa': [-96.7, 40.58, -90.1, 43.5],
    'Bangladesh1kmSubset': [78.9, 23.2, 96.1, 29.3]
}


def parameter_mapped(name, pft_array, bplut = BPLUT):
    '''
    Given a BPLUT parameter and a PFT array, returns an array with the
    corresponding parameter values for each PFT code.

    Parameters
    ----------
    name : str
        The name of the BPLUT parameter
    pft_array : numpy.ndarray
        Array of any size or shape with numeric elements corresponding to
        PFT codes
    bplut : dict
        (Optional) A collection of BPLUT parameters

    Returns
    -------
    numpy.ndarray
        Array of same size, shape as pft_array but with parameter values
        in place of PFT codes
    '''
    param = bplut[name]
    # Basically, index the <name> array, in PFT order, by PFT numeric codes
    return np.asarray(param)[np.ravel(pft_array)].reshape(pft_array.shape)


def restore_bplut(csv_file_path, version_id = None):
    '''
    Translates a BPLUT CSV file to a Python internal representation
    (OrderedDict instance).

    Parameters
    ----------
    csv_file_path : str
        File path to the CSV representation of the BPLUT
    version_id : str
        (Optional) Version identifier for the BPLUT

    Returns
    -------
    OrderedDict
    '''
    header = ('LC_index', 'LC_Label', 'model_code', 'NDVItoFPAR_scale',
        'NDVItoFPAR_offset', 'LUEmax', 'Tmin_min_K', 'Tmin_max_K',
        'VPD_min_Pa', 'VPD_max_Pa', 'SMrz_min', 'SMrz_max', 'FT_min',
        'FT_max', 'SMtop_min', 'SMtop_max', 'Tsoil_beta0', 'Tsoil_beta1',
        'Tsoil_beta2', 'fraut', 'fmet', 'fstr', 'kopt', 'kstr', 'kslw',
        'Nee_QA_Rank_min', 'Nee_QA_Rank_max', 'Nee_QA_Error_min',
        'Nee_QA_Error_max', 'Fpar_QA_Rank_min', 'Fpar_QA_Rank_max',
        'Fpar_QA_Error_min', 'Fpar_QA_Error_max', 'FtMethod_QA_mult',
        'FtAge_QA_Rank_min', 'FtAge_QA_Rank_max', 'FtAge_QA_Error_min',
        'FtAge_QA_Error_max', 'Par_QA_Error', 'Tmin_QA_Error',
        'Vpd_QA_Error', 'Smrz_QA_Error', 'Tsoil_QA_Error', 'Smtop_QA_Error')
    contents = []
    with open(csv_file_path, 'r') as stream:
        reader = csv.DictReader(
            filter(lambda row: row[0] != '#', stream), fieldnames = header)
        for row in reader:
            contents.append(row)

    result = BPLUT.copy()
    result['_version'] = 'UNKNOWN' if version_id is None else version_id
    result['LUE'] = np.array([[
        contents[p-1]['LUEmax'] if p in range(1, 9) else np.nan
        for p in range(0, 10)
    ]], dtype = np.float32).round(4)
    result['CUE'] = np.array([[
        (1 - float(contents[p-1]['fraut'])) if p in range(1, 9) else np.nan
        for p in range(0, 10)
    ]], dtype = np.float32).round(4)
    result['tmin'] = np.array([
        [contents[p-1]['Tmin_min_K'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['Tmin_max_K'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(1)
    result['vpd'] = np.array([
        [contents[p-1]['VPD_min_Pa'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['VPD_max_Pa'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(2)
    result['smrz'] = np.array([
        [contents[p-1]['SMrz_min'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['SMrz_max'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(1)
    result['smsf'] = np.array([
        [contents[p-1]['SMtop_min'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['SMtop_max'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(1)
    result['ft'] = np.array([
        [contents[p-1]['FT_min'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['FT_max'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(3)
    result['tsoil'] = np.array([[
        contents[p-1]['Tsoil_beta0'] if p in range(1, 9) else np.nan
        for p in range(0, 10)
    ]], dtype = np.float32).round(1)
    result['f_metabolic'] = np.array([[
        contents[p-1]['fmet'] if p in range(1, 9) else np.nan
        for p in range(0, 10)
    ]], dtype = np.float32).round(3)
    result['f_structural'] = np.array([[
        contents[p-1]['fstr'] if p in range(1, 9) else np.nan
        for p in range(0, 10)
    ]], dtype = np.float32).round(1)
    result['decay_rates'] = np.array([
        [contents[p-1]['kopt'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['kstr'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
        [contents[p-1]['kslw'] if p in range(1, 9) else np.nan
            for p in range(0, 10)],
    ], dtype = np.float32).round(3)
    # The "kstr" and "kslw" values are really the fraction of kopt
    #   assigned to the second and third pools
    result['decay_rates'][1,:] = np.multiply(
        result['decay_rates'][0,:], result['decay_rates'][1,:])
    result['decay_rates'][2,:] = np.multiply(
        result['decay_rates'][0,:], result['decay_rates'][2,:])
    return result



'''
Contains spatial reference system (SRS) definitions.
'''

EPSG = {
    6933: '''PROJCS["unnamed",
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            TOWGS84[0,0,0,0,0,0,0],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.0174532925199433,
            AUTHORITY["EPSG","9108"]],
        AUTHORITY["EPSG","4326"]],
    PROJECTION["Cylindrical_Equal_Area"],
    PARAMETER["standard_parallel_1",30],
    PARAMETER["central_meridian",0],
    PARAMETER["false_easting",0],
    PARAMETER["false_northing",0],
    UNIT["Meter",1],
    AUTHORITY["epsg","6933"]]''',
    6931: '''PROJCS["unnamed",
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            TOWGS84[0,0,0,0,0,0,0],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.0174532925199433,
            AUTHORITY["EPSG","9108"]],
        AUTHORITY["EPSG","4326"]],
    PROJECTION["Lambert_Azimuthal_Equal_Area"],
    PARAMETER["latitude_of_center",90],
    PARAMETER["longitude_of_center",0],
    PARAMETER["false_easting",0],
    PARAMETER["false_northing",0],
    UNIT["Meter",1],
    AUTHORITY["epsg","6931"]]''',
    6932: '''PROJCS["unnamed",
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            TOWGS84[0,0,0,0,0,0,0],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.0174532925199433,
            AUTHORITY["EPSG","9108"]],
        AUTHORITY["EPSG","4326"]],
    PROJECTION["Lambert_Azimuthal_Equal_Area"],
    PARAMETER["latitude_of_center",-90],
    PARAMETER["longitude_of_center",0],
    PARAMETER["false_easting",0],
    PARAMETER["false_northing",0],
    UNIT["Meter",1],
    AUTHORITY["epsg","6932"]]'''
}

SR_ORG = { # https://spatialreference.org/ref/sr-org/6974/
    6974: '''PROJCS["MODIS Sinusoidal",
    GEOGCS["WGS 84",
        DATUM["WGS_1984",
            SPHEROID["WGS 84",6378137,298.257223563,
                AUTHORITY["EPSG","7030"]],
            AUTHORITY["EPSG","6326"]],
        PRIMEM["Greenwich",0,
            AUTHORITY["EPSG","8901"]],
        UNIT["degree",0.01745329251994328,
            AUTHORITY["EPSG","9122"]],
        AUTHORITY["EPSG","4326"]],
    PROJECTION["Sinusoidal"],
    PARAMETER["false_easting",0.0],
    PARAMETER["false_northing",0.0],
    PARAMETER["central_meridian",0.0],
    PARAMETER["semi_major",6371007.181],
    PARAMETER["semi_minor",6371007.181],
    UNIT["m",1.0],
    AUTHORITY["SR-ORG","6974"]]''',
    6842: '''PROJCS["Sinusoidal",
    GEOGCS["GCS_Undefined",
        DATUM["Undefined",
            SPHEROID["User_Defined_Spheroid",6371007.181,0.0]],
        PRIMEM["Greenwich",0.0],
        UNIT["Degree",0.0174532925199433]],
    PROJECTION["Sinusoidal"],
    PARAMETER["False_Easting",0.0],
    PARAMETER["False_Northing",0.0],
    PARAMETER["Central_Meridian",0.0],
    UNIT["Meter",1.0]]
    '''
}



'''
Calibration procedure for SMAP Level-4 Carbon (L4C). For a list of all
commands, type:

    python main.py

**You must create a configuration JSON file before calibrating L4C.** This
file tells the calibration tool where on your file system the data files
are located. There is a template available in the directory:

    pyl4c/data/fixtures/files

The scratch data must be set-up before doing anything else:

    python main.py setup

Some commands can be chained together, for example, setting the PFT class to
calibrate is REQUIRED before most other commands; it can be set in one of
two ways:

    python main.py --pft=<pft> <command>
    python main.py pft <pft> <command>

Generally, the workflow for calibrating a single PFT is as follows:

    python main.py setup
    python main.py pft <pft> filter-preview gpp <window_size>
    python main.py pft <pft> filter         gpp <window_size>
    python main.py pft <pft> plot-gpp <driver>
    python main.py pft <pft> tune-gpp
    python main.py pft <pft> filter-preview reco <window_size>
    python main.py pft <pft> filter         reco <window_size>
    python main.py pft <pft> plot-reco <driver>
    python main.py pft <pft> tune-reco

Can optionally filter flux tower data for all PFTs:

    python main.py setup --reset
    python main.py filter-all gpp <window_size>
    python main.py filter-all reco <window_size>

**See the docstring on the CLI in this module for more information.**

Possible improvements:

- Replace parameter vectors with a
    `pyl4c.apps.calibration.ModelParameters` instance.
- Tower HDF5 file has upper-case field names (e.g., "APAR") while
    driver HDF5 file has lower-case field names (e.g., "par")
'''

import csv
import datetime
import itertools
import json
import os
import pickle
import h5py
import numpy as np
import pyl4c
from collections import Counter
from functools import partial
from scipy import signal
from matplotlib import pyplot
from pyl4c import pft_selector, suppress_warnings
from pyl4c.data.fixtures import PFT, restore_bplut
from pyl4c.science import arrhenius, climatology365
from pyl4c.stats import linear_constraint, detrend, rmsd
from pyl4c.lib.cli import ProgressBar
from pyl4c.apps.calibration import BPLUT, OPT_BOUNDS, GenericOptimization, cbar, report_fit_stats, solve_least_squares

CONFIG = os.path.join(os.path.dirname(pyl4c.__file__), 'data/files/config_calibration.json')

class CLI(object):
    '''
    Command line interface for calibrating L4C.

    To plot the response function of, e.g., Tmin, with (optional) suggested
    parameters (lower, upper bounds):

        python main.py pft <pft> plot-gpp <driver> [xmin, xmax]

    To optimize all of the GPP parameters:

        python main.py pft <pft> tune-gpp

    To optimize some of the GPP parameters, keeping named parameters fixed:

        python main.py pft <pft> tune-gpp --fixed="(LUE,ft0)"

    To optimize GPP with the best match for previous calibrations:

        python main.py pft <pft> tune-gpp --end="2014-12-31"

    To optimize GPP, first setting the initial value of a parameter:

        python main.py pft <pft> set <param> <value> tune-gpp

    Parameters
    ----------
    config : str
        Path to a configuration JSON file, which describes the file paths
        for calibration datasets
    pft : int
        The numeric code of the PFT to calibrate
    start : int
        (Optional) The numeric index making the start of a time series subset
        to use; if not provided, entire time series (in dataset) is used
    end : int
        (Optional) The numeric index making the end of a time series subset
        to use; if not provided, entire time series (in dataset) is used
    debug : bool
    use_legacy_pft : bool
        (Optional) True to use the L4C Nature Run v7.2 "legacy" PFT map
        ("lc_dom") for the PFT class assignments of each pixel (Default: True)
    '''
    _driver_bounds = {'apar': (2, np.inf)}
    _metadata = {
        'tmin': {'units': 'deg K'},
        'vpd': {'units': 'Pa'},
        'smrz': {'units': '%'},
        'smsf': {'units': '%'},
        'tsoil': {'units': 'deg K'},
    }
    _parameters = {
        'gpp': (
            'LUE', 'tmin0', 'tmin1', 'vpd0', 'vpd1', 'smrz0', 'smrz1', 'ft0'
        ),
        'reco': (
            'CUE', 'tsoil', 'smsf0', 'smsf1'
        )
    }
    _required_drivers = [
        'fpar', 'par', 'smrz', 'smsf', 'tsoil', 'tmin', 'vpd', 'tsurf'
    ]

    def __init__(
            self, config = CONFIG, pft = None, start = None, end = None,
            debug = True, use_legacy_pft = True):
        self._debug = debug
        self._nsites = 0
        self._nsteps = 0
        self._pft = pft
        self._pft_map = None
        self._site_weights = None
        self._time_end = None
        self._time_start = None
        self._use_legacy_pft = use_legacy_pft

        # Read in configuration file to determine file paths
        with open(config, 'r') as file:
            config_data = json.load(file)
        self._path_to_bplut = config_data['BPLUT_file']
        self._path_to_drivers = config_data['drivers_file']
        self._path_to_scratch = config_data['scratch_file']
        self._path_to_towers = config_data['towers_file']
        self._check() # Check for required keys in driver data

        # Allow users to specify start and end dates for the time series
        if start is not None or end is not None:
            with h5py.File(self._path_to_drivers, 'r') as hdf:
                time_series = [
                    datetime.datetime(y, m, d).strftime('%Y-%m-%d')
                    for y, m, d, _ in hdf['time'][:].tolist()
                ]
            if start is not None:
                self._time_start = time_series.index(start)
            if end is not None:
                self._time_end = time_series.index(end)

        # Creates the BPLUT store
        self._init_bplut()

        # Read in PFT map
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            self._nsites = hdf['state/PFT'].shape[0]
            self._nsteps = hdf['time'].shape[0]
            if self._use_legacy_pft:
                self._pft_map = hdf['legacy/lc_dom'][:].swapaxes(0, 1)
            else:
                self._pft_map = hdf['state/PFT'][:]

    @property
    def _is_setup(self):
        return os.path.exists(self._path_to_scratch)

    def _bounds(self, init_params, group, fixed = None, bounds = OPT_BOUNDS):
        'Defines bounds; optionally "fixes" parameters by fixing bounds'
        params = init_params
        if fixed is not None:
            params = [ # If the given parameter is in "fixed", restrict bounds
                None if self._parameters[group][i] in fixed else init_params[i]
                for i in range(0, len(self._parameters[group]))
            ]
        lower = []
        upper = []
        for i, p in enumerate(params):
            # This is a parameter to be optimized; use default bounds
            if p is not None:
                lower.append(bounds[group][0][i])
                upper.append(bounds[group][1][i])
            else:
                lower.append(init_params[i] - 1e-3)
                upper.append(init_params[i] + 1e-3)
        return (np.array(lower), np.array(upper))

    def _check(self):
        'Checks for all the required keys in the driver dataset'
        msg = 'Missing required driver: %s'
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            for key in self._required_drivers:
                assert key in hdf['drivers'].keys(), msg % key
            assert hdf['drivers/smsf'][:].max() > 1,\
                'SMSF data may not be in percent saturation units!'
            assert hdf['drivers/smrz'][:].max() > 1,\
                'SMRZ data may not be in percent saturation units!'

    @suppress_warnings
    def _climatology(self):
        'Computes a 365-day climatology for each driver'
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            time_series = [
                datetime.datetime(y, m, d)
                for y, m, d, _ in hdf['time'][:].tolist()
            ]
            with h5py.File(self._path_to_scratch, 'a') as target_hdf:
                for key in self._required_drivers:
                    if 'climatologies' in target_hdf.keys():
                        if key in target_hdf['climatologies'].keys():
                            continue # Skip to next

                    if self._debug:
                        print('Calculating %s climatology...' % key)
                    shp = (365, *hdf['drivers/%s' % key].shape[1:])
                    target_hdf.create_dataset(
                        'climatologies/%s' % key, shp,
                        dtype = hdf['drivers/%s' % key].dtype,
                        data = climatology365(
                            hdf['drivers/%s' % key][:], time_series))

    def _constrain(self, x, driver, coefs = None):
        'Converts a driver x into a multiple on [0, 1].'
        if driver == 'tsoil':
            return arrhenius(x, self.bplut['tsoil'][0,self._pft])
        # User can provide an M-element array-like
        if coefs is None:
            coefs = self.bplut[driver][:,self._pft].tolist()
        constraint = linear_constraint(coefs[0], coefs[1])
        if driver == 'vpd':
            # VPD mult. declines with increasing VPD, unlike other drivers
            constraint = linear_constraint(coefs[0], coefs[1], 'reversed')
        elif driver == 'ft':
            # FT has a binary response
            constraint = linear_constraint(coefs[0], coefs[1], 'binary')
        return constraint(x)

    def _filter(self, raw, size):
        'Apply a smoothing filter with zero phase offset'
        if size > 1:
            window = np.ones(size) / size
            return np.apply_along_axis(
                lambda x: signal.filtfilt(window, np.ones(1), x), 0, raw)
        else:
            return raw # Or, revert to the raw data

    def _init_bplut(self, labels = None):
        # Create the output BPLUT store; if there's already a BPLUT table
        #   in the scratch, it will be combined with the initial BPLUT,
        #   overwriting INTITIAL_BPLUT values in favor of the file BPLUT
        self.bplut = BPLUT(
            restore_bplut(self._path_to_bplut), labels = labels,
            hdf5_path = self._path_to_scratch)

    def _ramp(self, x, driver, step = 0.1, coefs = None):
        'Returns a ramp function over the domain of x; returns (domain, ramp)'
        domain = np.arange(np.nanmin(x), np.nanmax(x), step)
        ramp = self._constrain(domain, driver, coefs)
        return (domain, ramp)

    def _report(self, old_params, new_params, labels, title, prec = 2):
        'Prints a report on the updated (optimized) parameters'
        pad = max(len(l) for l in labels) + 1
        fmt_string = '-- {:<%d} {:>%d} [{:>%d}]' % (pad, 5 + prec, 7 + prec)
        print('%s parameters report, %s (PFT %d):' % (
            title, PFT[self._pft][0], self._pft))
        print((' {:>%d} {:>%d}' % (8 + pad + prec, 8 + prec))\
            .format('NEW', 'INITIAL'))
        for i, label in enumerate(labels):
            new = ('%%.%df' % prec) % new_params[i] if new_params[i] is not None else ''
            old = ('%%.%df' % prec) % old_params[i]
            print(fmt_string.format(('%s:' % label), new, old))

    @suppress_warnings
    def _report_fit(self, obs, pred, weights, verbose = True):
        'Reports RMSE and unbiased RMSE'
        return report_fit_stats(obs, pred, weights, verbose = verbose)

    @property
    def _sites(self):
        '''
        For a given PFT class, returns the tower sites, as rank indices, that
        represent that PFT. Exceptions are made according to the L4C
        calibration protocol, e.g., sites with any amount of
        Deciduous Needleleaf (DNF) in their 1-km subgrid are considered to
        represent the DNF PFT class.
        '''
        return pft_selector(self._pft_map, self._pft)

    def pft(self, pft):
        '''
        Sets the PFT class for the next calibration step.

        Parameters
        ----------
        pft : int
            The PFT class to use in calibration

        Returns
        -------
        CLI
        '''
        assert pft in range(1, 9), 'Unrecognized PFT class'
        self._pft = pft
        return self

    def filter(self, flux, size = 2):
        '''
        Filters the tower GPP or RECO flux data. Each time it is called, the
        raw (noisy) tower data is filtered; successive calls with the same
        arguments will only overwrite the filtered tower data.

        Parameters
        ----------
        flux : str
            "gpp" or "reco" for the corresponding tower dataset
        size : int
            Filter window size, in days
        '''
        assert any(self._sites), 'You must select a PFT class'
        with h5py.File(self._path_to_scratch, 'a') as hdf:
            assert flux.upper() in hdf['tower'].keys(),\
                'Field "tower/%s" not found' % flux.upper()
            # Tower fluxes are duplicated on the last axis, so pick one
            raw = hdf['tower/%s' % flux.upper()][:,self._sites,0]
            filtered = self._filter(raw, size)
            # Re-duplicate the 1-km subgrid on the last axis
            filtered = filtered[...,np.newaxis].repeat(81, axis = 2)
            hdf['tower/%s' % flux.upper()][:,self._sites,:] = filtered

    def filter_all(self, flux, size = 2):
        '''
        Filters the tower GPP or RECO flux data for ALL PFT classes.
        See also: filter().

        Parameters
        ----------
        flux : str
            "gpp" or "reco" for the corresponding tower dataset
        size : int
            Filter window size, in days
        '''
        with h5py.File(self._path_to_scratch, 'a') as hdf:
            assert flux.upper() in hdf['tower'].keys(),\
                'Field "tower/%s" not found' % flux.upper()
            # Tower fluxes are duplicated on the last axis, so pick one
            filtered = self._filter(
                hdf['tower/%s' % flux.upper()][...,0], size)
            # Re-duplicate the 1-km subgrid on the last axis
            filtered = filtered[...,np.newaxis].repeat(81, axis = 2)
            hdf['tower/%s' % flux.upper()][:] = filtered

    def filter_preview(self, flux, size = 2, seed = 9):
        '''
        Previews a filter for the tower GPP or RECO data.

        Parameters
        flux : str
            "gpp" or "reco" for the corresponding tower dataset
        size : int
            Filter window size, in days
        seed : int
            Random seed; change to select a different tower site
        '''
        assert self._sites.any(), 'You must select a PFT class'
        with h5py.File(self._path_to_scratch, 'r') as hdf:
            assert flux.upper() in hdf['tower'].keys(),\
                'Field "%s" not found' % flux.upper()
            x = hdf['tower/%s' % flux.upper()][:]

        # Get a random tower site matching the current PFT class
        np.random.seed(seed)
        idx = np.random.choice(np.argwhere(self._sites).flatten(), 1)[0]
        x0 = x[:,idx,0]
        x_filt = self._filter(x0, size)
        pyplot.plot(x0, c = 'lightsteelblue')
        pyplot.plot(x_filt, c = 'darkorange')
        pyplot.ylabel('Tower %s (g C m-2 day-1)' % flux.upper())
        pyplot.title('Window Size = %d days' % size)
        pyplot.show()

    def plot_gpp(
            self, driver, coefs = None, xlim = None, ylim = None, alpha = 0.1,
            marker = '.'):
        '''
        Using the current or optimized BPLUT coefficients, plots the GPP ramp
        function for a given driver. NOTE: Values where APAR < 2.0 are not
        shown.

        Parameters
        ----------
        driver : str
            Name of the driver to plot on the horizontal axis
        coefs : list or tuple or numpy.ndarray
            (Optional) array-like, Instead of using what's in the BPLUT,
            specify the exact parameters, e.g., [tmin0, tmin1]
        xlim : list or tuple
            (Optional) A 2-element sequence: The x-axis limits
        ylim : list or tuple
            (Optional) A 2-element sequence: The x-axis limits
        alpha : float
            (Optional) The alpha value (Default: 0.1)
        marker : str
            (Optional) The marker symbol (Default: ".")
        '''
        @suppress_warnings
        def empirical_lue(apar, gpp):
            # Mask low APAR values
            lower, _ = self._driver_bounds.get('apar', (0, None))
            apar = np.where(apar < lower, np.nan, apar)
            # Calculate empirical light-use efficiency: GPP/APAR
            return np.where(apar > 0, np.divide(gpp, apar), 0)

        assert self._is_setup, 'Must run setup first'
        np.seterr(invalid = 'ignore')
        # Read in GPP and APAR data
        if coefs is not None:
            assert hasattr(coefs, 'index') and not hasattr(coefs, 'title'),\
                "Argument --coefs expects a list [values,] with NO spaces"
        assert driver in ('tmin', 'vpd', 'smrz'),\
            'Requested driver "%s" cannot be plotted for GPP' % driver
        with h5py.File(self._path_to_scratch, 'r') as hdf:
            gpp = hdf['tower/GPP'][:,self._sites,:].mean(axis = 2)
            apar = hdf['APAR'][:,self._sites,:].mean(axis = 2)

        with h5py.File(self._path_to_drivers, 'r') as hdf:
            # Get the user-specified driver data
            x = hdf['drivers/%s' % driver][:,self._sites]

        # Update plotting parameters
        lue = empirical_lue(apar, gpp)
        # Mask out negative LUE values and values with APAR<2
        pyplot.scatter(x, np.where(
            np.logical_or(lue == 0, apar < 2), np.nan, lue),
            alpha = alpha, marker = marker)
        a, b = self._ramp(x, driver)
        pyplot.plot(a, b * self.bplut['LUE'][:,self._pft], 'k-')
        if coefs is not None:
            pyplot.plot(*self._ramp(x, driver, coefs = coefs), 'r-')
        pyplot.xlabel('%s (%s)' % (driver, self._metadata[driver]['units']))
        pyplot.ylabel('GPP/APAR (g C MJ-1 d-1)')
        if xlim is not None:
            pyplot.xlim(xlim[0], xlim[1])
        if ylim is not None:
            pyplot.ylim(ylim[0], ylim[1])
        pyplot.title(
            '%s (PFT %d): GPP Response to "%s"' % (
                PFT[self._pft][0], self._pft, driver))
        pyplot.show()

    def plot_reco(
            self, driver, coefs = None, q_rh = 75, q_k = 50, xlim = None,
            ylim = None, alpha = 0.1, marker = '.'):
        '''
        Using the current or optimized BPLUT coefficients, plots the RECO ramp
        function for a given driver. The ramp function is shown on a plot of
        RH/Cbar, which is equivalent to Kmult (as Cbar is an upper quantile of
        the RH/Kmult distribution).

        Parameters
        ----------
        driver : str
            Name of the driver to plot on the horizontal axis
        coefs : list or tuple or numpy.ndarray
            (Optional) array-like, Instead of using what's in the BPLUT,
            specify the exact parameters, e.g., `[tmin0, tmin1]`
        q_rh : int
            Additional arguments to `pyl4c.apps.calibration.cbar()`
        q_k : int
            Additional arguments to `pyl4c.apps.calibration.cbar()`
        ylim : list or tuple
            (Optional) A 2-element sequence: The x-axis limits
        alpha : float
            (Optional) The alpha value (Default: 0.1)
        marker : str
            (Optional) The marker symbol (Default: ".")
        '''
        assert self._is_setup, 'Must run setup first'
        assert driver in ('tsoil', 'smsf'),\
            'Requested driver "%s" cannot be plotted for RECO' % driver
        np.seterr(invalid = 'ignore')
        with h5py.File(self._path_to_scratch, 'r') as hdf:
            gpp = hdf['tower/GPP'][:,self._sites,:].mean(axis = 2)
            reco = hdf['tower/RECO'][:,self._sites,:].mean(axis = 2)

        with h5py.File(self._path_to_drivers, 'r') as hdf:
            # Get the user-specified driver data
            tsoil = hdf['drivers/tsoil'][:,self._sites]
            smsf = hdf['drivers/smsf'][:,self._sites]

        f_smsf = linear_constraint(*self.bplut['smsf'][:,self._pft])
        k_mult = f_smsf(smsf) * arrhenius(tsoil, self.bplut['tsoil'][0,self._pft])
        # Calculate RH as (RECO - RA)
        rh = reco - ((1 - self.bplut['CUE'][0,self._pft]) * gpp)
        # Set negative RH values to zero
        rh = np.where(suppress_warnings(np.less)(rh, 0), 0, rh)
        cbar0 = suppress_warnings(cbar)(rh, k_mult, q_rh, q_k)
        gpp = reco = None

        # Update plotting parameters
        pyplot.scatter( # Plot RH/Cbar against either Tsoil or SMSF
            tsoil if driver == 'tsoil' else smsf,
            suppress_warnings(np.divide)(rh, cbar0),
            alpha = alpha, marker = marker)

        if driver == 'tsoil':
            domain = np.arange(tsoil.min(), tsoil.max(), 0.1)
            pyplot.plot(domain,
                arrhenius(domain, self.bplut['tsoil'][0,self._pft]), 'k-')
        elif driver == 'smsf':
            pyplot.plot(*self._ramp(smsf, driver), 'k-')

        if coefs is not None:
            if driver == 'tsoil':
                pyplot.plot(domain, arrhenius(domain, *coefs), 'r-')
            elif driver == 'smsf':
                pyplot.plot(*self._ramp(smsf, driver, coefs = coefs), 'r-')

        pyplot.xlabel('%s (%s)' % (driver, self._metadata[driver]['units']))
        pyplot.ylabel('RH/Cbar')
        if xlim is not None:
            pyplot.xlim(xlim[0], xlim[1])
        if ylim is not None:
            pyplot.ylim(ylim[0], ylim[1])
        pyplot.title(
            '%s (PFT %d): RECO Response to "%s"' % (
                PFT[self._pft][0], self._pft, driver))
        pyplot.show()

    def reset(self):
        '''
        Clears the current BPLUT in the scratch data, wiping out any
        calibrated parameters. Subsequent calibration will then use the
        reference (file) BPLUT for initial values.
        '''
        with h5py.File(self._path_to_scratch, 'a') as hdf:
            del hdf['BPLUT']
        print('Deleted BPLUT in the scratch file')

    def set(self, parameter, value):
        '''
        Sets the named parameter to the given value for the specified PFT
        class. This updates the initial parameters, affecting any subsequent
        optimization.

        Parameters
        ----------
        parameter : str
            Name of the parameter to bet set
        value : int or float
            Value of the named parameter to be set

        Returns
        -------
        CLI
        '''
        # Update the BPLUT in memory but NOT the file BPLUT (this is temporary)
        self.bplut.update(self._pft, (value,), (parameter,), flush = False)
        return self

    def setup(self, reset = False):
        '''
        Compute freeze/thaw, copy tower data, calculate site weights, and
        compute a 365-day climatology for each driver.

        Important considerations:
            1) Tower GPP and RECO values <0 are masked;
            2) Tower GPP is masked where APAR is <0.1 MJ m-2 d-1;
            3) Freeze/Thaw status is determiend as "frozen" (0) if the surface
               skin temperature is below 273.15 K, "thawed" (1) otherwise

        Parameters
        ----------
        reset : bool
            True to automatically delete the scratch file
        '''
        np.seterr(invalid = 'ignore')
        if reset:
            os.remove(self._path_to_scratch)
            print('Deleted scratch data.')

        # Open the scratch file
        target_hdf = h5py.File(self._path_to_scratch, 'a')
        # Copy tower flux datasets
        with h5py.File(self._path_to_towers, 'r') as hdf:
            for key in ('GPP', 'RECO', 'NEE'):
                arr = hdf[key][:]
                arr = arr[...,np.newaxis].repeat(81, axis = 2)
                # Mask out negative GPP and RECO values
                if key != 'NEE':
                    arr = np.where(np.less(arr, 0), np.nan, arr)
                target_hdf.create_dataset(
                    'tower/%s' % key, arr.shape, arr.dtype, arr)

        # Read in Tsurf, 9-km grid indices, and PAR + fPAR
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            tsurf = hdf['drivers/tsurf'][:]
            idx = hdf['coords/grid_9km_idx'][:].round(0).astype(np.int16)
            # Make conformable PAR array at 1-km scale
            par = hdf['drivers/par'][:]
            par = par[...,np.newaxis].repeat(81, axis = 2)
            apar = np.multiply(hdf['drivers/fpar'][:], par)
            par = None

        # Compute freeze/thaw
        target_hdf.create_dataset( # Frozen = 0, Thawed = 1
            'ft', tsurf.shape, tsurf.dtype, np.where(tsurf <= 273.15, 0, 1))
        tsurf = None
        # Find duplicate (shared) 9-km cells, calculate site weights
        uid = ['%d%d' % (r, c) for r, c in idx.tolist()]
        target_hdf.create_dataset('site_weights', (1, self._nsites),
            np.float16, np.array([(1.0 / uid.count(x)) for x in uid]))
        # Copy APAR; screen tower GPP where APAR is low
        gpp = target_hdf['tower/GPP'][:]
        target_hdf.create_dataset('APAR', apar.shape, apar.dtype, apar)
        # With APAR below 0.1, tower GPP is unstable
        target_hdf['tower/GPP'][:] = np.where(apar < 0.1, np.nan, gpp)
        apar = gpp = None
        # Close the scratch data
        target_hdf.close()
        # Compute climatology
        self._climatology()

    def tune_gpp(
            self, fixed = None, optimize = True, nlopt = True, trials = 1):
        '''
        Optimizes GPP. The 9-km mean L4C GPP is fit to the tower-observed GPP
        using constrained, non-linear least-squares optimization.

        Parameters
        ----------
        fixed : tuple or list
            Zero or more parameters whose values should be fixed
            (i.e, NOT optimized)
        optimize : bool
            False to only report parameters and their fit statistics instead
            of optimizing (Default: True)
        nlopt : bool
            True to use the nlopt library for optimization (Default: True)
        '''
        def e_mult(params):
            # Calculate E_mult based on current parameters
            f_tmin = linear_constraint(params[1], params[2])
            f_vpd  = linear_constraint(params[3], params[4], 'reversed')
            f_smrz = linear_constraint(params[5], params[6])
            f_ft   = linear_constraint(params[7], 1.0, 'binary')
            tmin, vpd, smrz, ft = drivers # Unpack global "drivers"
            e = f_tmin(tmin) * f_vpd(vpd) * f_smrz(smrz) * f_ft(ft)
            return e[...,np.newaxis]

        def gpp(params):
            # Calculate GPP based on the provided BPLUT parameters
            return apar * params[0] * e_mult(params) # Global "apar"

        def residuals(params):
            # Objective function: Difference between tower GPP and L4C GPP
            gpp0 = gpp(params).mean(axis = 2)
            diff = np.subtract(gpp_tower, gpp0) # Global "gpp_tower"
            # Multiply by the tower weights
            return (weights * diff)[~np.isnan(diff)] # Global "weights"

        assert self._is_setup, 'Must run setup first'
        # Compile the initial parameters for optimization, based on current
        #   BPLUT; they are: lue, tmin0, tmin1, vpd0, vpd1, smrz0, smrz1, ft0
        if fixed is not None:
            # In case a single parameter name is passed, wrap in a list
            fixed = [fixed] if hasattr(fixed, 'title') else fixed
            assert all(p in self._parameters['gpp'] for p in fixed),\
                'Arguments to "fixed" should be in: [%s]' % ', '.join(self._parameters['gpp'])
        init_params = [self.bplut['LUE'][0,self._pft]]
        for field in ('tmin', 'vpd', 'smrz', 'ft'):
            init_params.extend(self.bplut[field][:,self._pft].tolist())
        init_params.pop() # Remove FT upper limit (it is fixed at 1.0)

        # Read in data, with optional subsetting of the time axis
        t0 = self._time_start if self._time_start is not None else 0
        t1 = self._time_end if self._time_end is not None else self._nsteps
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            drivers = []
            for field in ('tmin', 'vpd', 'smrz'):
                drivers.append(hdf['drivers/%s' % field][t0:t1][:,self._sites])

        with h5py.File(self._path_to_scratch, 'r') as hdf:
            drivers.append(hdf['ft'][t0:t1][:,self._sites])
            apar = hdf['APAR'][t0:t1][:,self._sites,:]
            gpp_tower = hdf['tower/GPP'][t0:t1][:,self._sites,:].mean(axis = 2)
            weights = hdf['site_weights'][:][:,self._sites]

        # L4C drivers should have no NaNs, based on how they were sourced;
        #   BUT this is not the case for fPAR, which does have NaNs; not sure
        #   why I would have ever asserted it doesn't
        # assert np.all(~np.isnan(apar)), 'Unexpected NaNs in APAR'
        for arr in drivers:
            assert np.all(~np.isnan(arr)), 'Unexpected NaNs'

        # Get bounds for the parameter search
        bounds = self._bounds(init_params, 'gpp', fixed)
        params = []
        params0 = []
        scores = []
        param_space = np.linspace(bounds[0], bounds[1], 100)
        for t in range(0, trials):
            # If multiple trials, randomize the initial parameter values
            #   and score the model in each trial
            if trials > 1:
                p = param_space.shape[1] # Number of parameters
                idx = np.random.randint(0, param_space.shape[0], p)
                init_params = param_space[idx,np.arange(0, p)]
                params0.append(init_params)
            if optimize and not nlopt:
                # Apply constrained, non-linear least-squares optimization
                print('Solving...')
                solution = solve_least_squares(
                    residuals, init_params, labels = self._parameters['gpp'],
                    bounds = self._bounds(init_params, 'gpp', fixed),
                    loss = 'arctan')
                fitted = solution.x.tolist()
                print(solution.message)
            elif optimize and nlopt:
                opt = GenericOptimization(residuals, bounds,
                    step_size = (0.01, 0.2, 0.2, 1, 1, 0.1, 0.1, 0.01))
                fitted = opt.solve(init_params)
            else:
                fitted = [None for i in range(0, len(init_params))]
            # Record the found solution and its goodness-of-fit score
            params.append(fitted)
            _, rmse_score, _, _ = self._report_fit(
                gpp_tower,
                gpp(fitted if optimize else init_params).mean(axis = 2),
                weights, verbose = False)
            print('[%s/%s] RMSE score of last trial: %.3f' % (
                str(t + 1).zfill(2), str(trials).zfill(2), rmse_score))
            scores.append(rmse_score)

        # Select the fit params with the best score
        if trials > 1:
            fitted = params[np.argmin(scores)]
            init_params = params0[np.argmin(scores)]
        # Generate and print a report, update the BPLUT parameters
        self._report(
            init_params, fitted, self._parameters['gpp'], 'GPP Optimization')
        self._report_fit(
            gpp_tower, gpp(fitted if optimize else init_params).mean(axis = 2),
            weights)
        if optimize:
            self.bplut.update(self._pft, fitted, self._parameters['gpp'])

    def tune_reco(
            self, q_rh = 75, q_k = 50, fixed = None, optimize = True,
            nlopt = True, trials = 1):
        '''
        Optimizes RECO. The 9-km mean L4C RECO is fit to the tower-observed
        RECO using constrained, non-linear least-squares optimization.
        Considerations:
            1) Negative RH values (i.e., NPP > RECO) are set to zero.

        Parameters
        ----------
        q_rh : int
            The percentile of RH/Kmult to use in calculating Cbar
        q_k : int
            The percentile of Kmult below which RH/Kmult values are masked
        fixed : tuple or list
            Zero or more parameters whose values should be fixed
            (i.e, NOT optimized)
        optimize : bool
            False to only report parameters and their fit statistics instead
            of optimizing (Default: True)
        nlopt : bool
            True to use the nlopt library for optimization (Default: True)
        trials : int
            Number of searches of the parameter space to perform; if >1,
            initial parameters are randomized for each trial
        '''
        def k_mult(params):
            # Calculate K_mult based on current parameters
            f_tsoil = partial(arrhenius, beta0 = params[1])
            f_smsf  = linear_constraint(params[2], params[3])
            tsoil, smsf = drivers # Unpack global "drivers"
            return f_tsoil(tsoil) * f_smsf(smsf)

        @suppress_warnings
        def reco(params):
            # Calculate RH as (RECO - RA) or (RECO - (faut * GPP));
            #   globals "reco_tower", "gpp_tower"
            ra = ((1 - params[0]) * gpp_tower)
            rh = reco_tower - ra
            rh = np.where(rh < 0, 0, rh) # Mask out negative RH values
            # Compute Cbar with globals "q_rh" and "q_k"
            kmult0 = k_mult(params)
            cbar0 = cbar(rh, kmult0, q_rh, q_k)
            return ra + (kmult0 * cbar0)

        def residuals(params):
            # Objective function: Difference between tower RECO and L4C RECO
            reco0 = reco(params)
            diff = np.subtract(reco_tower, reco0) # Global "reco_tower"
            missing = np.logical_or(np.isnan(reco_tower), np.isnan(reco0))
            # Multiply by the tower weights
            return (weights * diff)[~missing] # Global "weights"

        assert self._is_setup, 'Must run setup first'
        assert q_rh >= 0 and q_rh <= 100 and q_k >= 0 and q_k <= 100,\
            'Invalid setting for "q_rh" or "q_k" parameters'
        if fixed is not None:
            assert all(p in self._parameters['reco'] for p in fixed),\
                'Arguments to "fixed" should be in: [%s]' % ', '.join(self._parameters['reco'])
        init_params = [self.bplut['CUE'][0,self._pft]]
        for field in ('tsoil', 'smsf'):
            init_params.extend(self.bplut[field][:,self._pft].tolist())

        # Read in data, with optional subsetting of the time axis
        t0 = self._time_start if self._time_start is not None else 0
        t1 = self._time_end if self._time_end is not None else self._nsteps
        with h5py.File(self._path_to_drivers, 'r') as hdf:
            drivers = [
                hdf['drivers/%s' % field][t0:t1][:,self._sites]
                for field in ('tsoil', 'smsf')
            ]
        with h5py.File(self._path_to_scratch, 'r') as hdf:
            gpp_tower = hdf['tower/GPP'][t0:t1][:,self._sites,:].mean(axis = 2)
            reco_tower = hdf['tower/RECO'][t0:t1][:,self._sites,:].mean(axis = 2)
            weights = hdf['site_weights'][:,self._sites]

        # L4C drivers should have no NaNs, based on how they were sourced
        for arr in drivers:
            assert np.all(~np.isnan(arr)), 'Unexpected NaNs'

        # Get bounds for the parameter search
        bounds = self._bounds(init_params, 'reco', fixed)
        params = []
        params0 = []
        scores = []
        param_space = np.linspace(bounds[0], bounds[1], 100)
        for t in range(0, trials):
            # If multiple trials, randomize the initial parameter values
            #   and score the model in each trial
            if trials > 1:
                p = param_space.shape[1] # Number of parameters
                idx = np.random.randint(0, param_space.shape[0], p)
                init_params = param_space[idx,np.arange(0, p)]
                params0.append(init_params)
            if optimize and not nlopt:
                # Apply constrained, non-linear least-squares optimization
                solution = solve_least_squares(
                    residuals, init_params, labels = self._parameters['reco'],
                    bounds = bounds, loss = 'arctan')
                fitted = solution.x.tolist()
                message = solution.message
            elif optimize and nlopt:
                opt = GenericOptimization(residuals, bounds,
                    step_size = (0.01, 1, 0.1, 0.1))
                fitted = opt.solve(init_params)
                message = 'Success'
            else:
                fitted = [None for i in range(0, len(init_params))]
            # Record the found solution and its goodness-of-fit score
            params.append(fitted)
            _, rmse_score, _, _ = self._report_fit(
                reco_tower, reco(fitted if optimize else init_params),
                weights, verbose = False)
            print('[%s/%s] RMSE score of last trial: %.3f' % (
                str(t + 1).zfill(2), str(trials).zfill(2), rmse_score))
            scores.append(rmse_score)

        # Select the fit params with the best score
        if trials > 1:
            fitted = params[np.argmin(scores)]
            init_params = params0[np.argmin(scores)]
        # Generate and print a report, update the BPLUT parameters
        self._report(
            init_params, fitted, self._parameters['reco'], 'RECO Optimization')
        self._report_fit(
            reco_tower, reco(fitted if optimize else init_params), weights)
        if optimize:
            self.bplut.update(self._pft, fitted, self._parameters['reco'])


if __name__ == '__main__':
    import fire
    fire.Fire(CLI)



'''
Various statistical functions. Note that many have an `add_intercept` argument
and that this is True by default, which means the X matrix will have a column
of ones added for the calculation of the intercept. If you are already using
a design/ model matrix with an intercept term, be sure to set
`add_intercept = False`.
'''

from collections import Counter
import math
import numpy as np

summary_statistics = {
    'boxplot': {
        'nanmin': np.nanmin,
        'nanmax': np.nanmax,
        'nanmedian': np.nanmedian,
        'q1': lambda x: np.percentile(x[~np.isnan(x)], 25),
        'q2': lambda x: np.percentile(x[~np.isnan(x)], 75)
    }
}

def detrend(y, x = None, fill = False):
    '''
    Detrends a 1D series (linearly). Equivalently, returns the residuals
    of an OLS regression where the (transpose of the) design matrix is:

        1 1 1 1 ... 1
        0 1 2 3 ... N

    This removes the linear (straight line) component, e.g., of a time
    series with equal-size time steps.

    Parameters
    ----------
    y : numpy.ndarray
        The 1D series to detrend
    x : numpy.ndarray
        (Optional) The linear series describing the trend; usually a series
        of consecutive integers, e.g.: 1, 2, 3, ...
    fill : bool
        (Optional) True to fill NaNs with mean value (Default: False)

    Returns
    -------
    numpy.ndarray
        The detrended y values
    '''
    assert y.ndim == 1, 'Series to detrend must be a 1D vector'
    n, m = (y.size, 1)
    if x is not None:
        if x.ndim == 2:
            n, m = x.shape

    # Return all NaNs if the input array is all NaNs
    if np.all(np.isnan(y)):
        return np.repeat(np.nan, n, axis = 0)

    # Optionally, fill NaNs with the mean
    if fill:
        y = y.copy()
        nan_mask = np.isnan(y)
        y[np.isnan(y)] = np.nanmean(y)

    x = np.arange(0, n) if x is None else x
    beta = ols(x, y, add_intercept = True)
    yhat = np.hstack(
        (np.ones((n,)).reshape((n, 1)), x.reshape((n, m)))) @ beta
    if fill:
        return np.where(nan_mask, np.nan, np.subtract(y, yhat))
    return np.subtract(y, yhat)


def entropy(seq, base = 2):
    '''
    Returns the Shannon entropy of an input sequence of elements. Default
    units are bits but other units can be returned by changing the base.
    All calculations are with base-2 logarithms; change in base is done
    through multiplying by the constant factor `log_b(a)`` to change from
    base `a` to base `b`. Adapted from [1].

    NOTE: An estimate of the upper limit or "optimal" entropy for base `b`
    with `N` possible symbols can be obtained:

            math.log(n, b)

    1. http://rosettacode.org/wiki/Entropy#Python

    Parameters
    ----------
    seq : list or tuple or str
        Sequence of elements over which entropy is calculated
    base : int or float
        Base of the output units, e.g., 2 for "bits," e (2.718...) for "nats,"
        and 10 for "bans" (Default: 2)

    Returns
    -------
    float
    '''
    # Trivial case, all symbols are the same
    if np.all(np.equal(seq[0], seq)):
        return 0.0 # Avoid a "-0.0" return
    p, lns = Counter(seq), float(len(seq))
    e = -sum( count / lns * np.log2(count / lns) for count in p.values())
    return e if base == 2 else e * math.log(2, base)


def harmonic_ols(x, y, period = 12):
    r'''
    Returns OLS estimates for harmonic series of X, i.e., the matrix `A` of
    the equation `y = Ax + b` is a linear combination of sines and cosines.
    $$
    y_{i,t} = \alpha_i +
    \beta_{i,1}\, \mathrm{cos}\left(\frac{2\pi}{T}x_{i,t}\right) +
    \beta_{i,2}\,\mathrm{sin}\left(\frac{2\pi}{T}x_{i,t}\right) + \varepsilon_{i,t}
    $$

    Parameters
    ----------
    x : numpy.ndarray
        The independent variable, must be 1D
    y : numpy.ndarray
        The dependent variable, must be 1D

    Returns
    -------
    numpy.ndarray
        The solution to the least-squares problem
    '''
    assert x.ndim == 1, 'Array x must be 1D'
    x0 = x.reshape((x.shape[0], 1))
    # Create initial design/ model matrix (without intercept), as this is
    #   augmented by ols()
    a = ((2 * np.pi) / period) # Harmonic coefficient
    xm = np.hstack((x0, np.cos(a * x0), np.sin(a * x0)))
    return ols(xm, y, add_intercept = True)


def linear_constraint(xmin, xmax, form = None):
    '''
    Returns a linear ramp function, for deriving a value on [0, 1] from
    an input value `x`:

        if x >= xmax:
            return 1
        if x <= xmin:
            return 0
        return (x - xmin) / (xmax - xmin)

    Parameters
    ----------
    xmin : int or float
        Lower bound of the linear ramp function
    xmax : int or float
        Upper bound of the linear ramp function
    form : str
        Type of ramp function: "reversed" decreases as x increases;
        "binary" returns xmax when x == 1; default (None) is increasing
        as x increases.

    Returns
    -------
    function
    '''
    assert form == 'binary' or np.any(xmax >= xmin),\
        'xmax must be greater than/ equal to xmin'
    if form == 'reversed':
        return lambda x: np.where(x >= xmax, 0,
            np.where(x < xmin, 1, 1 - np.divide(
                x - xmin, xmax - xmin)))
    if form == 'binary':
        return lambda x: np.where(x == 1, xmax, xmin)
    return lambda x: np.where(x >= xmax, 1,
        np.where(x < xmin, 0, np.divide(x - xmin, xmax - xmin)))


def ols(x, y, add_intercept = True, use_qr_decomp = True):
    '''
    Returns ordinary least squares (OLS) estimates for X. If X is univariate
    (1D), returns the slope of the line between Y and X as well as the
    "y-intercept" or the intersection of this line with the vertical axis.

    Parameters
    ----------
    x : numpy.ndarray
        The independent variable(s); should N x M where M is the number of
        variables
    y : numpy.ndarray
        The dependent variable, must be 1D
    add_intercept : bool
        True to add a y-intercept term (Default: True)
    use_qr_decomp : bool
        True to use QR decomposition to obtain solution (Default: True)

    Returns
    -------
    numpy.ndarray
        The solution to the least-squares problem
    '''
    assert y.ndim == 1, 'Array y must be 1D'
    n = x.shape[0] # Num. of samples
    m = 1 if x.ndim == 1 else x.shape[1] # Num. of covariates
    # Create design/ model matrix
    xm = x.reshape((n, m)) # Without y-intercept term, unless...
    if add_intercept:
        xm = np.hstack((np.ones((n,)).reshape((n, 1)), x.reshape((n, m))))
    if xm.shape[1] > n:
        raise ValueError('System of equations is rank deficient')
    # Generally better to use QR decomposition to obtain \hat{\beta}
    if use_qr_decomp:
        q, r = np.linalg.qr(xm)
        fit = np.linalg.inv(r) @ q.T @ y
    else:
        fit = np.linalg.inv(xm.T @ xm) @ xm.T @ y
    return fit


def ols_variance(x, y, beta = None, add_intercept = True):
    r'''
    Returns the unbiased estimate of OLS model variance:
    $$
    SSE / (n - p)
    $$

    Where:
    $$
    SSE = (y - X\beta )' (y - X\beta )
    $$

    SSE is known as the sum of squared errors of prediction and is equivalent
    to the residual sum of squares (RSS).

    Parameters
    ----------
    x : numpy.ndarray
        The independent variable(s); should be N x M where M is the number of
        variables
    y : numpy.ndarray
        The dependent variable, must be 1D
    beta : numpy.ndarray
        (Optional) Coefficient estimates, an M-dimensional vector
    add_intercept : bool
        True to add a y-intercept term (Default: True)

    Returns
    -------
    float
        The sum-of-squared-errors of prediction
    '''
    n = x.shape[0] # Num. of samples
    p = 1 if x.ndim == 1 else x.shape[1] # Num. of covariates
    p += 1 if add_intercept else 0
    # Alternatively, use xm and beta calculated in sum_of_squares()...
    # return ((y - xm @ beta).T @ (y - xm @ beta)) / (n - p)
    return sum_of_squares(x, y, beta, add_intercept, which = 'sse') / (n - p)


def rmsd(x1, x2, n = None, weights = None):
    r'''
    Returns the root mean-squared deviation (RMSD) between two continuously
    varying random quantities:
    $$
    RMSD(\hat{x}, x) = \sqrt{n^{-1}\sum_i^N (\hat{x}_i - x_i)^2}
    $$

    Where `N` (or `n`) is the number of samples (e.g., model cells or time
    steps).

    NOTE: `NoData` should be filled with `np.nan` prior to calling this
    function; it is assumed that both vectors have the same missingness.

    Parameters
    ----------
    x1 : numpy.ndarray
        A 1D or 2D numeric vector
    x2 : numpy.ndarray
        A 1D or 2D numeric vector
    n : int
        (Optional) The number of samples, for normalizing; if not provided,
        calculated as the number of non-NaN samples
    weights : numpy.ndarray
        Weights array of a shape that can be broadcast to match both x1 and x2

    Returns
    -------
    float
    '''
    assert isinstance(n, int) or n is None, 'Argument "n" must be of integer type'
    assert x1.ndim <= 2 and x2.ndim <= 2, 'No support for more than 2 axes'
    diffs = np.subtract(x1, x2)
    # Optionally weight the residuals first
    if weights is not None:
        if weights.ndim == 1:
            weights = weights.reshape((1, *weights.shape))
        diffs = diffs * weights
    diffs = diffs[~np.isnan(diffs)]
    n = n if n is not None else diffs.shape[0]
    return np.sqrt(np.divide(np.sum(np.power(diffs, 2)), n))


def sum_of_squares(x, y, beta = None, add_intercept = True, which = 'ssr'):
    '''
    Calculates the sum-of-squared errors, sum-of-squared regressors, or
    total sum-of-squared errors based on the observed and predicted responses.
    Will calculate sum-of-squares for univariate linear regression if vector
    "x" is the independent variable and "y" is the dependent variable.

    Parameters
    ----------
    x : numpy.ndarray
        Observed response or independent variable;
    y : numpy.ndarray
        Predicted response or dependent variable;
    beta : numpy.ndarray
        (Optional) To avoid re-fitting (or fitting incorrectly) the OLS
        regression, can provide the coefficients of the OLS regression
    add_intercept : bool
        True to add an intercept term when fitting OLS regression
    which : str
        Which sum-of-squares quantity to calculate (Default: 'ssr');
        should be one of: ('sse', 'ssr', 'sst')

    Returns
    -------
    float
    '''
    assert y.ndim == 1, 'Array y must be 1D'
    if beta is None:
        beta = ols(x, y, add_intercept)
    n = x.shape[0] # Num. of samples
    m = p = 1 if x.ndim == 1 else x.shape[1] # Num. of covariates
    # Create design/ model matrix
    xm = x.reshape((n, m)) # Without y-intercept term, unless...
    if add_intercept:
        p += 1
        xm = np.hstack((np.ones((n,)).reshape((n, 1)), x.reshape((n, m))))

    yhat = xm @ beta # Calculate predicted values
    if which == 'sse':
        # Alternatively: ((y - xm @ beta).T @ (y - xm @ beta))
        return np.sum(np.power(np.subtract(y, yhat), 2))

    elif which == 'ssr':
        return np.sum(np.power(np.subtract(yhat, np.nanmean(y)), 2))

    elif which == 'sst':
        if add_intercept:
            return np.sum(np.power(np.subtract(y, np.nanmean(y)), 2))
        else:
            return np.sum(np.power(y, 2))

    else:
        raise NotImplementedError('Argument "which" not understood')


def t_statistic(x, y, beta_hat, beta = 0, add_intercept = True):
    r'''
    Calculates t-statistics for each of the predictors of beta_hat; by
    default, the true value of beta is assumed to be zero (i.e., null
    hypothesis is true). Calculated as:
    $$
    T_{n-p} = \frac{\hat{\beta }_j - \beta_j}{\mathrm{s.e.}(\hat{\beta }_j)}\sim t_{n-p}
    $$

    Parameters
    ----------
    x : numpy.ndarray
        The independent variable(s); should be N x M where M is the number
        of variables
    y : numpy.ndarray
        The dependent variable, must be 1D
    beta_hat : numpy.ndarray
        The estimated value(s) for beta
    beta : numpy.ndarray
        (Optional) The true value(s) for beta
    add_intercept : bool
        True to add a y-intercept term (Default: True)

    Returns
    -------
    float
    '''
    assert x.ndim <= 2, 'Array x must be 1D or 2D only'
    model_var = ols_variance(x, y, beta_hat, add_intercept = add_intercept) # Var(beta)
    beta_hat_se = np.sqrt(np.diag(np.linalg.inv(x.T @ x) * model_var))
    return np.subtract(beta_hat, beta) / beta_hat_se



'''
Convenience functions for working with SMAP L4C data in HDF5 arrays.
NOTE: All of the functions beginning with `get_` require access to ancillary
data files.
'''

import csv
import glob
import io
import os
import h5py
import numpy as np
from functools import partial
from concurrent.futures import ProcessPoolExecutor
from pyl4c import haversine
from pyl4c.data.fixtures import ANCILLARY_DATA_PATHS, HDF_PATHS, EASE2_GRID_PARAMS, SUBSETS_BBOX
from pyl4c.ease2 import ease2_coords, ease2_from_wgs84, ease2_search_radius

class MockL4CGranule(h5py.File):
    '''
    A mock for a typical L4C Ops granule. Should have all the right data
    fields at the right sizes for a typical L4C Ops granule read in as
    an h5py.File instance.

    Parameters
    ----------
    file : file
        (Optional) The File object to write; defaults to an in-memory
        `BytesIO` instance
    grid : str
        (Optional) The EASE-Grid 2.0 size: "M01" or "M09"
    mode : str
        (Optional) The file access mode (Default: "w")
    coords : tuple
        (Optional) 2D coordinate arrays to set on the granule's
        "GEO/longitude" and "GEO/latitude" fields
    data : dict
        (Optional) If provided, should be a Dictionary with keys for each
        desired dataset (e.g., "NEE" or "SOC") and any correctly shaped
        2D array as values
    pft_mean_fields : tuple or list
        Sequence of field names (e.g., `['RH', 'GPP']`) for which the PFT mean
        fields (e.g., `RH/rh_pft1_mean`) should be created
    '''
    def __init__(
            self, file = None, grid = 'M09', mode = 'w', coords = None,
            data = None, pft_mean_fields = list()):
        # Although file = io.BytesIO() in the function signature, above,
        #   should accomplish the same thing, it generates an OSError that
        #   cannot be reproduced in interactive mode unless the "file"
        #   is explicitly set here and io.BytesIO() is *not* invoked above
        if file is None:
            file = io.BytesIO()

        super(MockL4CGranule, self).__init__(file, mode)

        # Create a field for each L4C variable
        shp = EASE2_GRID_PARAMS[grid]['shape']
        for field in ('NEE', 'GPP', 'RH', 'SOC'):
            key = '%s/%s_mean' % (field, field.lower()) # e.g., "SOC/soc_mean"
            if data is None:
                init_data = np.ones(shp) * np.nan
            else:
                # Generate a NaN-valued array if no data provided
                init_data = data.get(key, np.ones(shp) * np.nan)

            # Create, e.g., "SOC/soc_mean" field
            self.create_dataset(key, shp, dtype = 'float32', data = init_data)
            # Optionally, create, e.g., "SOC/soc_pft1_mean" field
            if field in pft_mean_fields:
                for pft in range(1, 9):
                    key = '%s/%s_pft%d_mean' % (field, field.lower(), pft)
                    init_data = data.get(key, np.ones(shp) * np.nan)
                    self.create_dataset(
                        key, shp, dtype = 'float32', data = init_data)

        if coords is None:
            x_coords, y_coords = ease2_coords(grid = 'M09', in_1d = False)
        else:
            x_coords, y_coords = coords
            assert x_coords.ndim == 2 and y_coords.ndim == 2, 'Must provide 2D coordinate arrays'
        self.create_dataset('GEO/longitude', shp, dtype = 'float32',
            data = x_coords)
        self.create_dataset('GEO/latitude', shp, dtype = 'float32',
            data = y_coords)


def composite(
        *arrays, reducer = 'mean', target_band = 0, nodata = -9999.0,
        dtype = np.float32, processes = 1):
    '''
    Composites multiple raster arrays in a single band; is extremely fast when
    used with a Process Pool, but does not support normalization and will not
    composite more than one band at a time.

    NOTE: This function does NOT calculate memory use and will NOT guard
    against memory overflow.

    Parameters
    ----------
    arrays : *numpy.ndarray
        The input arrays to composite
    reducer : str
        The name of the reducer function, either: "median", "min", "max",
        "mean", or "std"
    target_band : int
        The index of the band to composite
    nodata : int or float
        The NoData value to ignore in compositing
    dtype : type
        The data type to enforce in the output
    processes : int
        (Optional) Number of processes to use

    Returns
    -------
    numpy.ndarray
    '''
    if reducer not in ('sum', 'median', 'min', 'max', 'mean', 'std'):
        raise ValueError('Invalid reducer name')

    assert all([isinstance(a, np.ndarray) for a in arrays]), 'Every element of "arrays" must be an instance of np.ndarray'
    shp = arrays[0].shape
    assert all(map(lambda x: x == shp, [r.shape for r in arrays])), 'Raster arrays must have the same shape'
    # For single-band arrays...
    if arrays[0].ndim < 3:
        shp = (1, shp[0], shp[1])
        arrays = list(map(lambda r: r.reshape(shp), arrays))

    # Stack the arrays in a continuous, single-band "tapestry" using
    #   vstack(), then cut out the arrays concatenated in this way into
    #   separate bands using reshape()
    b = target_band
    stack = np.vstack(list(map(
        lambda r: r[b,...].reshape((1, shp[1], shp[2])), arrays)))
    # Insert nan into NoData locations
    stack = np.where(stack == nodata, np.nan, stack)
    # Find the reducer function based on its name (should be a "nan" version)
    reducer_func = partial(getattr(np, 'nan%s' % reducer), axis = 0)
    # Avoid traceback masking and memory errors with ProcessPoolExecutor
    #   when only a single process is requested
    if processes == 1:
        all_results = reducer_func(stack)
    else:
        # Get index ranges for each process to work on
        work = partition(stack, num_processes = processes, axis = 1)
        with ProcessPoolExecutor(max_workers = processes) as executor:
            all_results = executor.map(
                reducer_func, [stack[:,i:j] for i, j in work])

    # Stack each reduced band (and reshape to multi-band image)
    result = np.concatenate(list(all_results), axis = 0)\
        .reshape((1, shp[1], shp[2]))
    return np.where(np.isnan(result), nodata, result)


def composite_hdf(hdf_file_glob, field, subset_id = 'CONUS', **kwargs):
    '''
    A convenience wrapper for `composite()`, enabling its use with HDF5 files
    specified by path names. Accepts additional `**kwargs` as keywoard arguments
    to `pyl4c.utils.composite()`.

    NOTE: This function does NOT calculate memory use and will NOT guard
    against memory overflow.

    Parameters
    ----------
    hdf_file_glob : str
        A file path expression for glob.glob()
    field : str
        The hierarchical path to a variable within each HDF5 file
    subset_id : str
        A well-known identifier for a bounding-box subset

    Returns
    -------
    numpy.ndarray
    '''
    assert subset_id in SUBSETS_BBOX.keys()
    assert len(glob.glob(hdf_file_glob)) > 0, 'No files found matching GLOB regular expression'
    arrays = []
    for path in glob.glob(hdf_file_glob):
        with h5py.File(path, 'r') as hdf:
            assert field in hdf.keys(), 'Variable "%s" not found in HDF5 file' % field
            arrays.append(subset(hdf, field, subset_id = subset_id)[0])
    return composite(*arrays, **kwargs)


def get_ease2_coords(grid, in_1d = True):
    '''
    DEPRECATED: Use `pyl4c.ease2.ease2_coords()` instead.
    Returns a tuple of 1D arrays, the X- and Y-coordinates of the desired
    EASE-Grid 2.0, from reading the ancillary data file.

    Parameters
    ----------
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.
    in_1d : bool
        True to return only 1D arrays, avoiding redundant rows/ columns
        (Default)

    Returns
    -------
    tuple
        X- and Y-coordinate arrays: (numpy.ndarray, numpy.ndarray)
    '''
    return ease2_coords(grid, in_1d)


def get_ease2_slice_idx(grid, subset_id):
    '''
    Returns a tuple `((xmin, xmax), (ymin, ymax))` of the indices that can be
    used to slice a corresponding EASE-Grid 2.0 in order to extract the
    desired bounding box.

    Parameters
    ----------
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.
    subset_id : str
        Keyword designating the desired subset and its corresponding bounding
        box

    Returns
    -------
    tuple
        Tuple of tuples: `((xmin, xmax), (ymin, ymax))`
    '''
    assert grid in EASE2_GRID_PARAMS.keys(),\
        'Could not understand grid argument; must be one of: %s' % ', '.join(EASE2_GRID_PARAMS.keys())
    x_coords, y_coords = ease2_coords(grid, in_1d = True)
    return get_slice_idx_by_bbox(x_coords, y_coords, subset_id)


def get_ease2_slice_offsets(grid, subset_id):
    '''
    Returns the X and Y offsets of an EASE-Grid 2.0 subset, so that an output
    raster array can be aligned properly; see
    `pyl4c.spatial.array_to_raster()`.

    Parameters
    ----------
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.
    subset_id : str
        Keyword designating the desired subset and its corresponding bounding box

    Returns
    -------
    tuple
        The X- and Y-offset `(xoff, yoff)`
    '''
    assert grid in EASE2_GRID_PARAMS.keys(),\
        'Could not understand grid argument; must be one of: %s' % ', '.join(EASE2_GRID_PARAMS.keys())
    slice_idx = get_ease2_slice_idx(grid = grid, subset_id = subset_id)
    return (slice_idx[0][0], slice_idx[1][0])


def get_pft_array(grid, subset_id = None, subset_bbox = None, slice_idx = None):
    '''
    Returns an array of PFT codes on the given EASE-Grid 2.0 for the entire
    globe or a specified subset.

    Arguments
    ---------
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.
    subset_id : str
        (Optional) Instead of `subset_bbox`, can provide keyword designating
        the desired subset area
    subset_bbox : tuple
        (Optional) Instead of `subset_id`, can provide any arbitrary bounding
        box, i.e., sequence of coordinates: `<xmin>, <ymin>, <xmax>, <ymax>`;
        these should be in WGS84 (longitude-latitude) coordinates
    slice_idx : list
        (Optional) Sequence of row-column slice indices to use, for subsetting
        the array; provide as: `[(row1, row2), (col1, col2)]` where the array
        is to be subset as: `array[row1:row2,col1:col2]`

    Returns
    -------
    numpy.ndarray
    '''
    assert grid in EASE2_GRID_PARAMS.keys(),\
        'Could not understand grid argument; must be one of: %s' % ', '.join(EASE2_GRID_PARAMS.keys())
    msg = 'Expected only one of the following arguments: subset_id, subset_bbox, slice_idx'
    assert not all((subset_id is not None, subset_bbox is not None)), msg
    assert not all((subset_bbox is not None, slice_idx is not None)), msg
    m = int(grid[-1]) # Pop off the grid size in km
    anc_hdf_path = ANCILLARY_DATA_PATHS['smap_l4c_ancillary_data_file_path']
    field = ANCILLARY_DATA_PATHS['smap_l4c_%dkm_ancillary_data_lc_path' % m]
    xf = ANCILLARY_DATA_PATHS['smap_l4c_%dkm_ancillary_data_x_coord_path' % m]
    yf = ANCILLARY_DATA_PATHS['smap_l4c_%dkm_ancillary_data_y_coord_path' % m]
    with h5py.File(anc_hdf_path, 'r') as hdf:
        if subset_id is not None:
            pft_array, _, _ = subset(
                hdf, field, hdf[xf][0,:], hdf[yf][:,0], subset_id = subset_id)
            return pft_array
        elif subset_bbox is not None:
            pft_array, _, _ = subset(
                hdf, field, hdf[xf][0,:], hdf[yf][:,0], subset_bbox = subset_bbox)
            return pft_array
        elif slice_idx is not None:
            y0, y1 = slice_idx[0]
            x0, x1 = slice_idx[1]
            return hdf[field][y0:y1,x0:x1]
        else:
            return hdf[field][:]


def get_slice_idx_by_bbox(
        x_coords, y_coords, subset_id = None, subset_bbox = None):
    '''
    Returns a tuple `((xmin, xmax), (ymin, ymax))` of the indices that can be
    used to slice a corresponding EASE2 grid in order to extract the
    desired bounding box. The returned indices correspond to the approximate
    locations of the bounding box (defined in geographic space) in array
    index space; i.e., if `array` is the EASE2 grid of interest, then the
    following slice will extract an area within the bounding box (bbox):

        array[ymin:ymax, xmin:xmax]

    NOTE: Coordinate arrays (`x_coords`, `y_coords`) must be given in the same
    units as the BBOX definition in `SUBSET_BBOX`, i.e., decimal degrees.

    Parameters
    ----------
    x_coords : numpy.ndarray
        A 1D array of X coordinate values
    y_coords : numpy.ndarray
        A 1D array of Y coordinate values
    subset_id : str
        (Optional) Instead of subset_bbox, can provide keyword designating
        the desired subset area
    subset_bbox : tuple
        (Optional) Instead of subset_id, can provide any arbitrary bounding
        box, i.e., sequence of coordinates: `<xmin>, <ymin>, <xmax>, <ymax>`;
        these should be in WGS84 (longitude-latitude) coordinates
    '''
    if subset_id is not None and subset_bbox is not None:
        raise ValueError('Should provide only one argument: Either subset_id or subset_bbox')
    assert x_coords.ndim == 1 and y_coords.ndim == 1, 'Must provide 1D coordinate arrays only'

    bb = subset_bbox
    if subset_id is not None:
        bb = SUBSETS_BBOX[subset_id]

    # Get min, max indices of X coords within bbox; note that we +1 because
    #   ending slice indices are non-inclusive in Python
    if x_coords[-1] > x_coords[0]:
        # If X-coordinates are sorted smallest to largest...
        x_slice_idx = [
            np.where(x_coords >= bb[0])[0].min(),
            np.where(x_coords <= bb[2])[0].max() + 1
        ]
    else:
        x_slice_idx = [
            np.where(x_coords <= bb[0])[0].min(),
            np.where(x_coords >= bb[2])[0].max() + 1
        ]

    # Get min, max indices of Y coords within bbox
    if y_coords[-1] > y_coords[0]:
        # If Y-coordinates are sorted smallest to largest...
        y_slice_idx = [
            np.where(y_coords <= bb[3])[0].max() + 1,
            np.where(y_coords >= bb[1])[0].min()
        ]
    else:
        y_slice_idx = [
            np.where(y_coords <= bb[3])[0].min(),
            np.where(y_coords >= bb[1])[0].max() + 1
        ]

    x_slice_idx.sort() # Necessary b/c slicing is always small:large number
    y_slice_idx.sort()
    return (tuple(x_slice_idx), tuple(y_slice_idx))


def get_xy_coords(hdf_or_nc, in_1d = True):
    '''
    Returns a tuple (longitude, latitude) where the elements are coordinate
    arrays of longitude and latitude. These are needed for, e.g., plotting
    the geophysical data on a global geographic grid. This is convenience
    function for extracting the longitude-latitude coordinates based on the
    filename and our knowledge of where these data are stored. NOTE: This may
    seem like a hack, but is the easiest solution to the fundamental problem
    of inconsistent variable paths; inconsistent naming between dataset IDs
    and filenames; and incomplete documentation of each within the HDF5 file.

    Parameters
    ----------
    hdf_or_nc : h4py.File or netcdf4.Dataset
        Either: an HDF5 file / h5py.File object OR a NetCDF file
    in_1d : bool
        True (Default) to return 1D arrays; if False, returns 2D arrays where
        the coordinates are duplicated along one axis
    '''
    try:
        # HDF5 files
        if hasattr(hdf_or_nc, 'keys') and hasattr(hdf_or_nc, 'filename'):
            if 'GPP' in hdf_or_nc.keys() and 'NEE' in hdf_or_nc.keys():
                d = HDF_PATHS['SPL4CMDL']['4'] # NOTE: We assume Version 4
            elif 'Geophysical_Data' in hdf_or_nc.keys():
                if not 'sm_surface' in hdf_or_nc['Geophysical_Data'].keys():
                    raise ValueError()
                d = HDF_PATHS['SPL4SMGP']['4']
            else:
                raise ValueError()
            x = hdf_or_nc[d['longitude']]
            y = hdf_or_nc[d['latitude']]
        # NetCDF files
        else:
            assert hasattr(hdf_or_nc, 'variables'), 'Assumed NetCDF file has no variables'
            if 'lon' in hdf_or_nc.variables and 'lat' in hdf_or_nc.variables:
                x = hdf_or_nc.variables['lon']
                y = hdf_or_nc.variables['lat']

    except ValueError:
        # Let's assume we know what the X-Y coordinate array keys are
        if 'cell_lon' in hdf_or_nc.keys() and 'cell_lat' in hdf_or_nc.keys():
            x = hdf_or_nc['cell_lon']
            y = hdf_or_nc['cell_lat']
        else:
            raise NotImplementedError('The filename was not recognized as a product with known longitude-latitude data')

    x = x[0,:] if in_1d and len(x.shape) > 1 else x[:]
    y = y[:,0] if in_1d and len(y.shape) > 1 else y[:]
    return (x, y)


def index(array, indices):
    '''
    Fast array indexing by subsetting the array first. If there are multiple
    indices, anywhere in the array, that need to be returned, this provides
    a speed-up by creating a smaller array to index.

    Parameters
    ----------
    array : numpy.ndarray
        Array containing data at (row, column) indices
    indices : tuple or list
        2-element sequence of (row, column) indices, in order

    Returns
    -------
    list
        List of indexed values
    '''
    assert len(indices) == 2,\
        'Must provide 2-element sequence of (row coordinates, column coordinates)'
    assert len(indices[0]) == len(indices[1]),\
        'Array indices must be the same length!'
    assert array.ndim <= 2, 'Array must have 2 or fewer axes'
    midx = np.ravel_multi_index(indices, array.shape)
    return array.ravel()[midx]


def partition_generator(n_elements, n_parts = 1):
    '''
    A Generator that yields slice indices for equal-sized partitions of a
    1D array or sequence. See also: `partition()`.

    Parameters
    ----------
    n_elements : int
        The number of elements in array/ sequence
    n_parts : int
        The number of partitions desired

    Returns
    -------
    generator
    '''
    p = 0
    parts = np.linspace(0, n_elements, n_parts + 1, dtype = int)
    while p < n_parts:
        yield (
            int(parts[p]),
            # If it is the final index, add 1
            int(parts[p+1] if (p != n_parts - 1) else (parts[p+1] + 1))
        )
        p += 1


def partition(array, num_processes, axis = 0):
    '''
    Creates index ranges for partitioning an array to work on over multiple
    processes.

    Parameters
    ----------
    array : numpy.ndarray
        The 2-dimensional array to partition
    num_processes : int
        The number of processes desired
    axis : int
        The axis to break apart into chunks

    Returns
    -------
    list
    '''
    N = array.shape[axis]
    return list(partition_generator(N, num_processes))


def subset(
        hdf_or_nc, field, x_coords = None, y_coords = None, subset_id = None,
        subset_bbox = None):
    '''
    Returns a subset array from the HDF, for the desired variable, where the
    array corresponds to an area defined by a known bounding box, e.g., the
    continental United States (CONUS).

    NOTE: `x_coords` and `y_coords` (hierarchical paths) will be inferred from
    the filename if not provided at all.

    Parameters
    ----------
    hdf_or_nc : h5py.File or netcdf4.Dataset
        Either: an HDF5 file / h5py.File object OR a NetCDF file
    field : str
        Hierarchical path to the desired variable
    x_coords : numpy.ndarray or str
        (Optional) 1D NumPy array of X coordinates OR hierarchical path to
        the variable representing X coordinates, e.g., longitude values
    y_coords : numpy.ndarray or str
        (Optional) 1D NumPy array of Y coordinates OR hierarchical path to
        the variable representing Y coordinates, e.g., latitude values
    subset_id : str
        (Optional) Instead of subset_bbox, can provide keyword designating
        the desired subset area
    subset_bbox : tuple or list
        (Optional) Instead of subset_id, can provide any arbitrary bounding
        box, i.e., sequence of coordinates: `<xmin>, <ymin>, <xmax>, <ymax>`;
        these should be in WGS84 (longitude-latitude) coordinates

    Returns
    -------
    tuple
        Tuple of: subset array, xoff, yoff; (numpy.ndarray, Int, Int)
    '''
    assert (subset_id is None and subset_bbox is not None) or (subset_id is not None and subset_bbox is None), 'Should provide only one argument: Either subset_id or subset_bbox'
    assert isinstance(hdf_or_nc, h5py.File) or hasattr(hdf_or_nc, 'variables'), 'An HDF5 or NetCDF file is required; cannot subset a stand-alone array'
    assert (x_coords is None and y_coords is None) or (isinstance(x_coords, str) and isinstance(y_coords, str)) or (isinstance(x_coords, np.ndarray) and isinstance(y_coords, np.ndarray)), 'The x_coords and y_coords arguments must have matching type'
    assert (not isinstance(x_coords, np.ndarray)) or (x_coords.ndim == 1), 'The x_coords and y_coords arguments must be 1D arrays, otherwise pass a String or None'

    # Check that we have a 2D array to work with
    if hasattr(hdf_or_nc, 'variables'):
        shp = hdf_or_nc.variables[field].shape
    else:
        assert field in hdf_or_nc.keys(), 'Field name "%s" not found' % field
        shp = hdf_or_nc[field].shape
    assert len(shp) == 2, 'HDF5 or NetCDF data array indexed by "%s" must be a 2D array' % field

    # If a hierarchical path to the X and Y coordinate variables was not
    #   given, then infer the paths from the filename
    if x_coords is None:
        x_coords, y_coords = get_xy_coords(hdf_or_nc, in_1d = True)
    if isinstance(x_coords, str):
        # NOTE: Only need first row of X-coordinate array, first column of
        #   Y-coordinate array, because they are duplicated thereafter
        if hasattr(hdf_or_nc, 'variables'):
            x_coords = hdf_or_nc.variables[x_coords][0,:] # NetCDF
            y_coords = hdf_or_nc.variables[y_coords][0,:]
        else:
            x_coords = hdf_or_nc[x_coords][0,:] # HDF5
            y_coords = hdf_or_nc[y_coords][:,0]

    x_idx, y_idx = get_slice_idx_by_bbox(
        x_coords, y_coords, subset_id, subset_bbox)
    xmin, xmax = x_idx # Unpack the slice range indexes
    ymin, ymax = y_idx
    if hasattr(hdf_or_nc, 'variables'):
        return (hdf_or_nc.variables[field][ymin:ymax, xmin:xmax], xmin, ymin)
    return (hdf_or_nc[field][ymin:ymax, xmin:xmax], xmin, ymin)


def sample(array, indices):
    '''
    Samples the value in an array at each (row, column) position in a sequence
    of row-column index pairs.

    Parameters
    ----------
    array : numpy.ndarray
    indices : tuple
        Tuple of (x, y) coordinate pairs; no z-level should be be provided
        (2D coordinate pairs only)

    Returns
    -------
    numpy.ndarray
    '''
    indices_by_axis = [
        int(i[0]) for i in indices], [int(i[1]) for i in indices
    ]
    # If row, column indices are not sorted, we'll need another list comprehension
    try:
        return array[indices_by_axis]
    except TypeError:
        return [array[idx[0], idx[1]] for idx in indices]


def summarize(
        data_array, summaries, scale = 1, data_mask = None, nodata = -9999):
    '''
    Calculates statistical summar[ies] of an input data array over all values.

    Parameters
    ----------
    data_array : numpy.ndarray
        Array of data values that we wish to summarize, e.g., gridded land
        surface temperatures
    summaries : dict
        A Dictionary of `{name: function}` where function is some vectorized
        function that acts over the values associated with a class label,
        including NaNs, and returns a single number
    scale : int or float
        Optional scaling parameter to apply to the input array values, e.g.,
        if the array values are (spatial) rates and should be scaled by the
        (equal) area of the grid cell
    nodata : int or float
        NoData or Fill value(s) to ignore; can pass a sequence of multiple
        values (Default: -9999)

    Returns
    -------
    dict
    '''
    assert data_array.ndim <= 2 or (data_array.ndim == 3 and data_array.shape[0] == 1), 'Can only work with 1-band raster arrays'
    if data_array.ndim == 3:
        data_array = data_array[0,...] # Unwrap 1-band raster arrays

    # Fill in NaN where there is NoData
    data_array = np.where(np.isin(data_array, nodata), np.nan, data_array)
    stats = dict([(k, None) for k in summaries.keys()])
    for stat_name, func in summaries.items():
        # NOTE: Runs faster if dtype of accumulator is *not* set
        stats[stat_name] = func(np.multiply(data_array, scale))
    return stats


def summarize_by_class(
        data_array, class_array, summaries, scale = 1, ignore = (0,),
        nodata = -9999):
    '''
    Calculates statistical summar[ies] of an input data array for each class
    label in an input class array.

    Parameters
    ----------
    data_array : numpy.ndarray
        Array of data values that we wish to summarize, e.g., gridded land
        surface temperatures
    class_array : numpy.ndarray
        Array of class labels that will be used to summarize input data,
        e.g., watersheds or ecoregions; should NOT contain any NaNs as this
        will cause the function to hang
    summaries : dict
        A Dictionary of `{name: function}` where function is some vectorized
        function that acts over the values associated with a class label,
        including NaNs, and returns a single number
    scale : int or float
        Optional scaling parameter to apply to the input array values, e.g.,
        if the array values are (spatial) rates and should be scaled by the
        (equal) area of the grid cell
    ignore : tuple
        Class labels (values in the class_array) to ignore
    nodata : int or float
        NoData or Fill value to ignore (Default: -9999)

    Returns
    -------
    dict
        A nested Python dictionary with a key-value pair for each class,
        where the value is another dictionary with a key-value pair for each
        summary statistic.
    '''
    assert data_array.ndim <= 2 or (data_array.ndim == 3 and data_array.shape[0] == 1), 'Can only work with 1-band raster arrays'
    if data_array.ndim == 3:
        data_array = data_array[0,...] # Unwrap 1-band raster arrays
    assert data_array.shape == class_array.shape, 'Input data_array does not match shape of the class_array'
    # Fill in NaN where there is NoData
    data_array = np.where(np.isin(data_array, nodata), np.nan, data_array)
    # In case None was passed to "ignore," replace with empty list
    ignore = ignore if ignore is not None else []
    assert not np.any(np.isnan(class_array)),\
        'Class array should not contain NaNs; use "ignore" argument instead'
    classes = set(np.unique(class_array[~np.isnan(class_array)])) # Create, e.g., {1: {}, 2: {}, ...}
    stats = dict([(k, dict()) for k in classes.difference(ignore)])
    for code in classes.difference(ignore):
        query = np.where(np.isin(class_array, code), data_array, np.nan)
        stats[code] = summarize(query, summaries, scale, nodata = nodata)
    return stats


def summarize_hdf_by_pft(
        hdf, field, summaries, scale = 1, pft_codes = range(1, 9),
        subset_id = 'CONUS', x_coords = None, y_coords = None,
        nodata = -9999):
    '''
    Calculates statistical summar[ies] of an input data array for each Plant
    Functional Type (PFT) class.

    Parameters
    ----------
    hdf : h5py.File
        The HDF5 file / h5py.File object
    field : str
        One of: "SOC", "NEE", "GPP", or "RH"
    summaries : dict
        A Dictionary of `{name: function}` where function is some vectorized
        function that acts over the values associated with a class label,
        including NaNs, and returns a single number
    scale : int or float
        Optional scaling parameter to apply to the input array values, e.g.,
        if the array values are (spatial) rates and should be scaled by the
        (equal) area of the grid cell
    pft_codes   The PFT codes to create summaries for; defaults to all
                of them, (i.e. Default: 1-8 inclusive)
    subset_id   Keyword designating the desired subset area
                (Default: CONUS)
    x_coords    (Optional) A 1D NumPy array of X coordinate values, used
                in subsetting and can be automatically discovered
    y_coords    (Optional) A 1D NumPy array of Y coordinate values, used
                in subsetting and can be automatically discovered
    nodata      NoData or Fill value to ignore (Default: -9999)
    '''
    assert field in ('SOC', 'NEE', 'GPP', 'RH'), 'Only possible to summarize one of: "SOC", "NEE", "GPP", or "RH"'
    assert (subset_id in SUBSETS_BBOX.keys()) or (subset_id is None), 'Named subset_id not a recognized geographic subset; see SUBSETS_BBOX in this module'
    # Warn user against using nansum() or sum() functions
    if any([k.rfind('sum') > 0 for k in summaries.keys()]):
        print('WARNING: Totals or sums may be biased low because subgrid heterogeneity is not recognized; use total_hdf_by_pft() instead')

    field_names = [
        '%s/%s_pft%d_mean' % (field, field.lower(), p) for p in pft_codes
    ]
    stats = dict([(p, dict()) for p in field_names])
    for name in field_names:
        if subset_id is not None:
            array, x, y = subset(
                hdf, name, x_coords, y_coords, subset_id)
        else:
            array = hdf[name][:]

        stats[name] = summarize(array, summaries, scale, nodata = nodata)

    return stats


def total_hdf_by_pft(
        hdf, counts, field, scale = 1, pft_codes = range(1, 9),
        max_count = 81, subset_id = 'CONUS', x_coords = None, y_coords = None,
        nodata = -9999):
    '''
    Calculates statistical summar[ies] of an input data array for each Plant
    Functional Type (PFT) class. The problem with calculating *totals* for
    PFT means is that we need to know the area proportion of a given PFT
    within each 9-km cell in order to correctly sum over that area. This
    function correctly scales the 9-km PFT mean value by the provided `scale`
    argument and then scales this value by the proportion of 1-km subgrid
    cells that match the given PFT. This produces accurate sums of spatial
    rates; i.e., the `scale` parameter converts a PFT mean to a (biased)
    total, then the `counts` are used to scale that total by the area
    proportion of the given PFT. The summary function used is np.nansum.

    Parameters
    ----------
    hdf : h5py.File
        The HDF5 file / h5py.File object
    counts : tuple or list
        A sequence of NumPy arrays, each array corresponding to the count of
        1-km subcells matching a certain PFT; these should be provided in
        order, i.e., first element is the count of PFT 1 subcells, second
        element for PFT 2, and on.
    field : str
        One of: "SOC", "NEE", "GPP", or "RH"
    scale : int or float
        Optional scaling parameter to apply to the input array values, e.g.,
        if the array values are (spatial) rates and should be scaled by the
        (equal) area of the grid cell
    pft_codes : tuple or list
        The PFT codes to create summaries for; defaults to all of them,
        (i.e. Default: 1-8 inclusive)
    max_count : int
        The maximum number of 1-km subcells that any given PFT can total
        within a 9-km area; no reason why this shouldn't be 81 (9 x 9),
        probably.
    subset_id : str
        (Optional) Instead of subset_bbox, can provide keyword designating
        the desired subset area
    x_coords : numpy.ndarray or str
        (Optional) 1D NumPy array of X coordinates OR hierarchical path to
        the variable representing X coordinates, e.g., longitude values
    y_coords : numpy.ndarray or str
        (Optional) 1D NumPy array of Y coordinates OR hierarchical path to
        the variable representing Y coordinates, e.g., latitude values
    nodata : int or float
        NoData or Fill value to ignore (Default: -9999)
    '''
    assert len(counts) == len(pft_codes),\
        'Length of counts and pft_codes must be equal'
    assert field in ('SOC', 'NEE', 'GPP', 'RH'),\
        'Only possible to summarize one of: "SOC", "NEE", "GPP", or "RH"'
    assert (subset_id in SUBSETS_BBOX.keys()) or (subset_id is None),\
        'Named subset_id not a recognized geographic subset; see SUBSETS_BBOX in this module'
    summaries = {'nansum': np.nansum}
    stats = dict()
    for i, pft in enumerate(pft_codes):
        # Get the name of the target field; set up summary stats dict()
        fieldname = '%s/%s_pft%d_mean' % (field, field.lower(), pft)
        stats[fieldname] = dict()

        if subset_id is not None:
            array, _, _ = subset(
                hdf, fieldname, x_coords, y_coords, subset_id)
        else:
            array = hdf[fieldname][:]

        assert array.shape == counts[i].shape, 'Counts array must be the same size as data array'
        # Scale the data array by the proportion of the subgrid cells that
        #   match the given PFT class
        stats[fieldname] = summarize(
            np.multiply( # NOTE: We scale data array ahead of time...
                np.divide(counts[i], max_count),
                np.multiply(array, scale)),
            # ...So scale must be fixed at 1
            summaries, scale = 1, nodata = nodata)

    return stats



'''
Tools for representing flux tower sites and working with flux tower data.

NOTE: Tower subgrid should accurately reflect the canonical EASE-Grid 2.0
nesting, i.e., the 1-km cells returned correspond to those nested under the
respective 9-km cell, which is apparent if we look at the subgrid of the
top-left cell:

    >>> TowerSite('test', ease2_to_wgs84((0, 0)), validate = False).subgrid
'''

import csv
import glob
import h5py
import numpy as np
from cached_property import cached_property
from pyl4c import haversine
from pyl4c.ease2 import ease2_search_radius, ease2_to_wgs84, ease2_from_wgs84
from pyl4c.utils import get_pft_array, index

class MetaTowerSite(type):
    '''
    Prototype for Tower classes.
    '''
    @property
    def pft_1km_map(cls):
        if getattr(cls, '_pft_1km_map', None) is None:
            cls._pft_1km_map = get_pft_array('M01')
        return cls._pft_1km_map

    @property
    def pft_9km_map(cls):
        if getattr(cls, '_pft_9km_map', None) is None:
            cls._pft_9km_map = get_pft_array('M09')
        return cls._pft_9km_map


class TowerSite(object, metaclass = MetaTowerSite):
    '''
    Represents an eddy covariance tower site.

    Parameters
    ----------
    id : str
        Unique name for the TowerSite
    coords : tuple | list
        The site's (longitude, latitude) coordinates
    validate : bool
        True to validate the PFT of any mapped EASE-Grid 2.0 cell
    '''
    valid_pft_range = range(1, 9)

    def __init__(self, id, coords, validate = True):
        assert hasattr(coords[0], 'real') and hasattr(coords[1], 'real'),\
            'Coordinates must be numbers, not strings or other types'
        self.id = id
        self.coords = coords
        self.idx_1km = ease2_from_wgs84(coords, 'M01', exact = True)
        self.idx_9km = ease2_from_wgs84(coords, 'M09', exact = True)
        self._subgrid_idx = None
        self._subgrid_pft = None
        self._validate = validate
        if validate:
            self.locate_ease2() # Fire it up!

    @cached_property
    def pft_9km(self):
        'The PFT class at 9-km scale'
        r, c = self.idx_9km
        return self.__class__.pft_9km_map[int(r), int(c)]

    @property
    def subgrid(self):
        'Returns a list of 1-km subgrid indices'
        if self._subgrid_idx is None:
            # A search radius of 4 pixels yields a width, height of 9 pixels
            self._subgrid_idx = ease2_search_radius(self.coords, 4, 'M01')
        return self._subgrid_idx

    @property
    def subgrid_pft(self):
        'Returns an 81-element array of the 1-km subgrid PFTs'
        if self._subgrid_pft is None:
            self._subgrid_pft = index(
                self.__class__.pft_1km_map,
                np.array(self.subgrid).swapaxes(0, 1).tolist())
        return np.array(self._subgrid_pft)

    def locate_ease2(self):
        '''
        Finds the EASE-Grid 2.0 row, column indices for this site.

        Returns
        -------
        list
            A list of tuples, each a 2-element sequence of (row, column)
            indices
        '''
        row, col = self.idx_9km
        pft = self.__class__.pft_9km_map[int(row), int(col)] # Expensive
        if self._validate and pft not in self.valid_pft_range:
            # Get the closest valid EASE-Grid 2.0 cell instead
            try:
                row, col = self.locate_ease2_alternatives(self.coords)
                # Chain forward and reverse transforms to get the grid
                #   cell indices in floating point
                row, col = ease2_from_wgs84(
                    ease2_to_wgs84((row, col)), 'M09', exact = True)
            except ValueError:
                raise ValueError('Could not locate suitable EASE-Grid 2.0 cell')

        # Update the 9-km grid coordinates, as needed
        self.idx_9km = row, col
        return self.idx_9km

    def locate_ease2_alternatives(self, coords, grid = 'M09'):
        '''
        Finds nearest EASE-Grid 2.0 cell w/ valid PFT up to 1 grid cell away.

        Parameters
        ----------
        coords : tuple or list
            2-element sequence of (longitude, latitude) coordinates
        grid : str
            EASE-Grid 2.0 cell size designation, e.g, "M01" or "M09"
            (Default: "M09")

        Returns
        -------
        list
            A list of tuples, each a 2-element sequence of (row, column)
            indices
        '''
        # Round the row, column indices of this point so as to check for
        #   collisions in haversine(), below
        row, col = ease2_from_wgs84(coords, grid, exact = False)
        indices = ease2_search_radius(coords, 1, grid)
        indices_pfts = index(
            self.__class__.pft_9km_map,
            np.array(indices).swapaxes(0, 1).tolist())
        options = [] # Make sure the choices also have valid PFTs
        for i, idx in enumerate(indices):
            if indices_pfts[i] in self.valid_pft_range:
                options.append(idx)
        dists = [ # Calculate great circle distance
            haversine(
                coords, ease2_to_wgs84((r,c), grid)
            ) for r, c in options if not (r == row and c == col)
        ]
        return options[np.argmin(dists)]


def consolidate_tower_data(
        file_glob, output_hdf5_path, n_sites = 356, n_steps = 6575,
        keys = ('nee', 'gpp', 'reco')):
    '''
    Combines the CSVs from multiple calibration tower sites into a single HDF5
    dataset, e.g.:

        consolidate_tower_data(
            "/anx_lagr2/laj/smap/fluxnet2015/merged/*_smapMergedCalVal.csv",
            "/anx_lagr3/arthur.endsley/SMAP_L4C/calibration/Fluxnet2015_LaThuile_tower_data_for_356_sites.h5")

    Parameters
    ----------
    file_glob : str
        GLOB expression targeting all input CSV files
    output_hdf5_path : str
        File path to output HDF5 file
    n_sites : int
        Number of tower sites
    n_steps : int
        Number of time steps
    keys : tuple
        Sequence of keys to extract from each CSV
    '''
    file_paths = glob.glob(file_glob)
    assert len(file_paths) == n_sites, 'Should be one file per tower site'
    all_records = []
    for path in file_paths:
        site_records = []
        with open(path, 'r') as stream:
            reader = csv.DictReader(stream)
            for each in reader:
                site_records.append([
                    np.float(each[k]) if each[k] != 'NaN' else np.nan
                    for k in keys
                ])

            all_records.append(site_records)
            site_records = None

    shp = (n_steps, n_sites) # Output a T x N array
    arr = np.array(all_records) # Becomes an (N x T x k) array
    all_records = None
    with h5py.File(output_hdf5_path, 'a') as hdf:
        for i, field in enumerate(keys):
            hdf.create_dataset(field, shp, dtype = np.float32,
                data = arr[...,i].swapaxes(0, 1))



'''
The L4C Science model, for computing a total carbon budget for specific
geographic point locations. This is a "point" version, similar to the Matlab
calibration code, that runs L4C for individual cells; it is NOT intended to
run L4C daily for the global domain.

The `L4CForwardProcessPoint` has the same model logic as the L4C operational
algorithm that is run at Goddard Space Flight Center. However, it does not
produce "bit-to-bit" identical results as L4C Ops, chiefly because of
differences in the spatial reference system (SRS) definitions used for
extracting and compiling driver data, but also because the initial soil
organic carbon (SOC) data does not match. This mismatch is hard to account
for; we've tried multiple SOC restart files but none of them match what
Goddard is using. It's possible this is also due entirely to the SRS mismatch.

Possible optimizations:

- Pre-compute the product of APAR and optimal LUE.
- Will break with L4C MDL (Ops) code, but it's more memory efficient
    to assign initial SOC pool sizes to t=0 and makes very (very) little
    difference in the end.
'''

import datetime
import numpy as np
import h5py
from pyl4c import suppress_warnings
from pyl4c.data.fixtures import restore_bplut
from pyl4c.science import arrhenius
from pyl4c.stats import linear_constraint
from pyl4c.lib.cli import ProgressBar
from pyl4c.apps.l4c import L4CConstants, L4CDrivers, L4CState, report
from pyl4c.apps.l4c.io import L4CStreamingInputDataset

class L4CForwardProcessPoint(object):
    '''
    Outputs of L4C include both State variables (soil organic carbon) and Flux
    variables (NEE, GPP, RH).

    NOTE: In `debug = True` mode, the Emult and Kmult diagnostic parameters
    are disaggregated and the individual environmental constraint multipliers
    are stored. Outside of debug mode, only Emult, Tmult, and Wmult are
    reported.

    Parameters
    ----------
    config : dict
        Dictionary of configuration parameters
    stream : bool
        True to use L4CStreamingInputDataset instead of reading in all driver
        data; this reduces memory use but increases I/O
    verbose : bool
        True to print all output to stdout
    debug : bool
        True to store additional diagnostic information from each time step
    '''
    DAYS_PER_YEAR = 365
    BOUNDS = {
        'apar': [0, np.inf],
        'vpd': [np.nan, np.nan],
        'ft': [0, 1],
        'smrz': [0, 100],
        'smsf': [0, 100],
        'tmin': [-200, 400],
        'tsoil': [-200, 400],
        'tsurf': [-200, 400],
    }
    CONSTANTS = {
        'tsoil_beta1': 66.02,
        'tsoil_beta2': 227.13,
        'tsurf_freeze-thaw_threshold_kelvin': 273.15
    }
    CONSTANTS_INDEX = (
        'CUE', 'LUE', 'f_metabolic', 'f_structural', 'decay_rates')
    # These MUST be listed in a fixed order...
    CONSTRAINTS_INDEX = ( # Constraints on GPP or RECO
        'tmin', 'vpd', 'ft', 'smrz', 'tsoil', 'smsf')
    DIAGNOSTICS_INDEX = (
        'f_tmin', 'f_vpd', 'f_ft', 'f_smrz', 'f_tsoil', 'f_smsf', 'apar')
    DRIVERS_INDEX = ('apar', 'tmin', 'vpd', 'ft', 'smrz', 'tsoil', 'smsf')
    FLUX_INDEX = ('gpp', 'rh', 'nee')
    STATE_INDEX = ('soc1', 'soc2', 'soc3', 'e_mult', 't_mult', 'w_mult')
    PFT_CODES = range(1, 10)
    REQUIRED_CONFIGURATION = ('bplut', 'inputs_file_path', 'site_count')

    def __init__(
            self, config, stream = True, use_legacy_pft = False,
            verbose = True, debug = False):
        self._check_configuration(config)
        self._bplut = config['bplut']
        self._config = config
        self._constraints = dict()
        self._debug = debug
        self._multilayer = False # Single soil layer
        self._streaming = stream
        self._t0 = 0 # Starting time index
        self._t1 = config['time_steps']
        self._time_idx = -1
        self._verbose = verbose
        self._use_legacy_pft = use_legacy_pft
        self.file_path = config['inputs_file_path']
        if verbose:
            print('NOTE: Running with BPLUT version %s' % self._bplut['_version'])
        with h5py.File(self.file_path, 'r') as hdf:
            self._setup(config, hdf)

    @property
    def config(self):
        return self._config

    @property
    def constants(self):
        return self._constants

    @property
    def drivers(self):
        return self._drivers # NOTE: No setter

    @property
    def fluxes(self):
        return self._fluxes

    @fluxes.setter
    def fluxes(self, new):
        self._fluxes.data = new

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new):
        self._state.data = new

    def _arrhenius(self, tsoil):
        'The Arrhenius equation for response of enzymes to (soil) temperature'
        beta1, beta2 = (self.CONSTANTS['tsoil_beta%d' % i] for i in (1, 2))
        return arrhenius(tsoil, self._constraints['tsoil'], beta1, beta2)

    def _load_constants(self):
        'Load arrays of constants for each PFT class, to speed up computation'
        for label in self.CONSTANTS_INDEX:
            self.constants.add(self._bplut[label], label)

    def _load_constraints(self, pft, drivers_constrained):
        'Creates [3 x N x 81] array of lower, upper bounds for ramp functions'
        # The third axis is the range (max - min), or difference between the
        #   first and second axes
        by_pft = pft.ravel()
        # Set invalid PFT codes to PFT 0
        by_pft = np.where(np.in1d(by_pft, self.PFT_CODES), by_pft, 0)
        shp = (2, self.config['site_count'], 81)

        # For each driver, propagate the BPLUT coefficients by PFT class
        for driver in drivers_constrained:
            # Tsoil requires a single coefficient; not lower, upper bounds
            if driver == 'tsoil':
                self._constraints['tsoil'] = np.apply_along_axis(
                    lambda p: self._bplut['tsoil'][0,p], 0, by_pft)\
                    .reshape((shp[1], shp[2]))
                continue

            # Get lower and upper bounds by PFT class
            self._constraints[driver] = np.apply_along_axis(
                lambda p: self._bplut[driver][:,p], 0, by_pft).reshape(shp)

    def _load_drivers(self, hdf, keys):
        'Load driver variables all at once from the HDF'
        new_drivers = []
        for key in keys:
            if key not in ('apar', 'ft'):
                assert key in hdf['drivers'].keys(),\
                    'Required driver "%s" not found in inputs HDF5' % key

            if key == 'apar':
                for k in ('fpar', 'par'):
                    assert k in hdf['drivers'].keys(),\
                        'Required driver "%s" not found in inputs HDF5' % k
                par = hdf['drivers/par'][self._t0:,...]
                par = par.reshape((*par.shape, 1)).repeat(81, axis = 2)
                # APAR is the product of fPAR and PAR
                new_drivers.append(
                    np.multiply(hdf['drivers/fpar'][self._t0:,...], par))
                continue

            if key == 'ft':
                # Use "ft" or "tsurf" depending on what's available to obtain
                #   a freeze-thaw record
                if 'ft' in hdf['drivers'].keys():
                    self._print('Using existing freeze-thaw driver data instead of "tsurf"')
                    new_drivers.append(hdf['drivers/ft'][self._t0:,...])
                    continue
                elif 'tsurf' in hdf['drivers'].keys():
                    k = self.CONSTANTS['tsurf_freeze-thaw_threshold_kelvin']
                    self._print('Calculating freeze-thaw condition using "tsurf" and cutoff of %f degrees K' % k)
                    new_drivers.append( # Frozen = 0, Thawed = 1
                        np.where(
                            hdf['drivers/tsurf'][self._t0:,...] <= k, 0, 1))
                    continue
                else:
                    raise ValueError('No freeze-thaw driver data found')

            # In all other cases
            new_drivers.append(hdf['drivers/%s' % key][self._t0:,...])

        return new_drivers

    @suppress_warnings
    def _load_state(self, hdf, keys):
        '''
        Load state variables (soil organic carbon) all at once from the HDF.
        An array large enough to hold model state is created ("new_state"),
        but also the initial state ("init_state") is created.
        '''
        shp = (1, self.config['time_steps'], self.config['site_count'], 81)
        init_state = []
        new_state = []
        for p, key in enumerate(keys):
            # Create an empty state array, allocated T time steps
            new_state.append(np.full(shp, np.nan))
            if key.startswith('soc'):
                arr = hdf['state/soil_organic_carbon'][p,...]
                # Filter out any NoData, which (should) only correspond to
                #   1-km subgrid pixels that are outside the PFT range [1, 8]
                init_state.append(
                    np.where(arr < 0, np.nan, arr).reshape((1, 1, *shp[2:])))
            else:
                if key not in self.DIAGNOSTICS_INDEX:
                    init_state.append(np.full((1, 1, *shp[2:]), np.nan))
        return (init_state, new_state)

    def _setup(self, config, hdf):
        'Load point site PFTs, state data, driver data'
        # Get the starting time index, if specified
        if 'start' in config.keys():
            if config['start'] is not None:
                try:
                    ts0 = datetime.datetime.strptime(
                        config['start'], '%Y-%m-%dT%H:%M:%S')
                except ValueError:
                    ts0 = datetime.datetime.strptime(
                        config['start'], '%Y-%m-%d')
                self._t0 = np.argwhere(
                    np.logical_and(np.logical_and(
                        hdf['time'][:,0] == ts0.year,
                        hdf['time'][:,1] == ts0.month),
                        hdf['time'][:,2] == ts0.day)
                    ).flatten().tolist().pop()
        if 'end' in config.keys():
            if config['end'] is not None:
                try:
                    ts1 = datetime.datetime.strptime(
                        config['end'], '%Y-%m-%dT%H:%M:%S')
                except ValueError:
                    ts1 = datetime.datetime.strptime(
                        config['end'], '%Y-%m-%d')
                self._t1 = np.argwhere(
                    np.logical_and(np.logical_and(
                        hdf['time'][:,0] == ts1.year,
                        hdf['time'][:,1] == ts1.month),
                        hdf['time'][:,2] == ts1.day)
                    ).flatten().tolist().pop()

        # And check that the correct number of steps were specified
        assert config['time_steps'] == (hdf['time'].shape[0] - self._t0)\
            or config['time_steps'] == (self._t1 - self._t0),\
            'Parameter "time_steps" does not match the number of time steps suggested by "start" parameter and the "time" field'

        self._print('Accessing state and drivers data...')
        if self._use_legacy_pft:
            self._pft = hdf['legacy/lc_dom'][:].swapaxes(0, 1)
        else:
            self._pft = hdf['state/PFT'][:]

        # Initialize containers for various datasets
        self._setup_data_storage(config, hdf)

        # Calculate daily litterfall based on the annual NPP sum
        self.constants.add(
            hdf['state/npp_sum'][:] / self.DAYS_PER_YEAR, 'litterfall')

        # SET STATE
        self._print('Loading state...')
        keys = list(self.STATE_INDEX)
        if self._debug:
            keys.extend(self.DIAGNOSTICS_INDEX)
        init_state, new_state = self._load_state(hdf, keys)
        if len(init_state) > 0:
            self.state_initial = np.concatenate(init_state, axis = 0)
        if len(new_state) > 0:
            self.state = np.concatenate(new_state, axis = 0)
        self.state.labels = keys

        # SET DRIVERS
        if not self._streaming:
            self._print('Loading drivers...')
            self._drivers = L4CDrivers(
                config['site_count'],
                self._load_drivers(hdf, self.DRIVERS_INDEX),
                labels = self.DRIVERS_INDEX)

        # LOAD CONSTANTS
        self._print('Loading constants...')
        self._load_constants()

        # INITIALIZE CONSTRAINT FUNCTIONS
        self._print('Creating linear constraint functions...')
        self._load_constraints(self._pft, self.CONSTRAINTS_INDEX)

    def _setup_data_storage(self, config, hdf):
        'Initialize containers for various datasets'
        site_names = hdf['site_id'][:].tolist()
        shp = (config['time_steps'], config['site_count'], 81)
        self._constants = L4CConstants(
            config['site_count'], self._pft, self.PFT_CODES)
        self._drivers = None # We set this only once, below
        self._fluxes = L4CState(
            config['site_count'], np.full((len(self.FLUX_INDEX), *shp), np.nan),
            self.FLUX_INDEX, axis_labels = [None, None, site_names, None])
        self._state = L4CState(config['site_count'],
            axis_labels = [None, None, site_names, None])

    def _print(self, message):
        'Only print to screen if in "verbose" mode'
        if self._verbose:
            print(message)

    def _check_configuration(self, config):
        'Checks for the existing of certain keys in the configuration file'
        for key in self.REQUIRED_CONFIGURATION:
            assert key in config.keys(), '%s not found' % key

    def constrain(self, x, driver):
        '''
        Returns a linear interpolated multiple based on a ramp function.
        Equivalent to a vectorized version of:

            if x >= xmax:
                return 1
            if x <= xmin:
                return 0
            return (x - xmin) / (xmax - xmin)

        Parameters
        ----------
        x : float
            Observed value, generally an (N x 81) array or an array with an
            (N x 81) or (N x 1) sub-space (e.g., soil moisture values that
            are constant over the 1-km grid have no 81 unique values), so they
            will be broadcast over the 81-pixel sub-grid when combined with
            constraint parameter array
        driver : str
            Name of the driver

        Returns
        -------
        float
        '''
        if driver == 'tsoil':
            return self._arrhenius(x)

        coefs = self._constraints[driver]
        if driver == 'vpd':
            # VPD mult. declines with increasing VPD, unlike other drivers
            return linear_constraint(coefs[0], coefs[1], form = 'reversed')(x)

        if driver == 'ft':
            # FT has a binary response
            return linear_constraint(coefs[0], coefs[1], form = 'binary')(x)

        return linear_constraint(coefs[0], coefs[1])(x)

    def gpp(self, drivers):
        '''
        Calculate GPP for a single time step.

        Parameters
        ----------
        drivers : list or tuple
            Nested sequence of (driver: `numpy.ndarray`, label: `str`) pairs

        Returns
        -------
        numpy.ndarray
        '''
        # Extract APAR, translate other drivers into environ. constraints
        apar, f_tmin, f_vpd, f_ft, f_smrz = [
            self.constrain(driver, label) if label != 'apar' else driver
            for driver, label in drivers
        ]
        e_mult = f_tmin * f_vpd * f_ft * f_smrz
        if self._debug:
            return (apar * self.constants.LUE * e_mult,
                (f_tmin, f_vpd, f_ft, f_smrz))
        return (apar * self.constants.LUE * e_mult, e_mult)

    def rh(self, state, drivers):
        '''
        Calculate RH for a single time step.

        Parameters
        ----------
        state : numpy.ndarray
            `(3 x N x M)` array of current SOC state in each pool
        drivers : list or tuple
            Nested sequence of (driver: `numpy.ndarray`, label: `str`) pairs

        Returns
        -------
        numpy.ndarray
        '''
        # Translate Tsoil and SMSF into environmental constraints on RH
        f_tsoil, f_smsf = [
            self.constrain(driver, label) for driver, label in drivers
        ]
        k_mult = (f_tsoil * f_smsf)[np.newaxis,...].repeat(3, axis = 0)
        # NOTE: These are true decay rates for 2nd and 3rd pools, so it
        #   is straightforward to multiply them against SOC
        rh = k_mult * self.constants.decay_rates * state
        # "the adjustment...to account for material transferred into the
        #   slow pool during humification" (Jones et al. 2017 TGARS, p.5);
        #   note that this is a loss FROM the "medium" (structural) pool,
        #   see tcfModFunc.c Lines 54-55
        rh[1,...] = rh[1,...] * (1 - self.constants.f_structural)
        # T_mult, W_mult same for each pool
        return (rh, (f_tsoil, f_smsf))

    def soc(self, rh, t = None):
        '''
        Calculate change in SOC for a single time step.

        Parameters
        ----------
        rh : numpy.ndarray
            `(3 x N x M)` array of RH at the current time step
        t : int
            Current time step; useful in subclasses but not used in the
            operational L4C algorithm (not used here)

        Returns
        -------
        numpy.ndarray
        '''
        # Change in SOC according to diff. eq. in Jones et al. (2017)
        litter = self.constants.litterfall
        dc1 = (litter * self.constants.f_metabolic) - rh[0,...]
        dc2 = (litter * (1 - self.constants.f_metabolic)) - rh[1,...]
        dc3 = (self.constants.f_structural * rh[1,...]) - rh[2,...]
        return (dc1, dc2, dc3)

    def run(self, steps = None, fields_gpp = None, fields_rh = None):
        '''
        A forward run in serial over multiple time steps; currently works in
        streaming mode ONLY.

        Parameters
        ----------
        steps : int
            Number of time steps to run or None to run through the end of the
            available time steps (exhaust driver data) (Default: `None`)
        fields_gpp : list or tuple or None
            (Optional) Sequence of field names that are used to drive the
            GPP model
        fields_rh : list or tuple or None
            (Optional) Sequence of field names that are used to drive the
            RH model
        '''
        @suppress_warnings
        def step(t, fields_gpp, fields_rh):
            'Calculate fluxes, new states for the next time step t'
            if t == 0:
                # Retrieve initial SOC pool sizes
                state = self.state_initial[0:3,0,...]
            else:
                # Retrieve SOC in each pool from prior step
                state = self.state.data[0:3,t-1,...]
            # Calculate fluxes, new states
            gpp, e_mult = self.gpp(
                zip(hdf.index(t + self._t0, *fields_gpp), fields_gpp))
            rh, k_mult = self.rh(
                state, zip(hdf.index(t + self._t0, *fields_rh), fields_rh))
            npp = gpp * self.constants.CUE
            d_soc = self.soc(rh, t + self._t0)
            # Record fluxes for this time step
            self.fluxes.update('gpp', t, gpp)
            self.fluxes.update('rh', t, rh.sum(axis = 0))
            self.fluxes.update('nee', t, rh.sum(axis = 0) - npp)
            # Record diagnostics at this time step
            if self._debug:
                # In debug mode, e_mult is not a single quantity but multiple;
                #   the order of e_mult_fields MUST match the order of the
                #   return signature for gpp()
                e_mult_fields = ('f_tmin', 'f_vpd', 'f_ft', 'f_smrz')
                if all(f in self.DIAGNOSTICS_INDEX for f in e_mult_fields):
                    for e, key in enumerate(self.DIAGNOSTICS_INDEX):
                        if key in e_mult_fields:
                            idx = e_mult_fields.index(key)
                            self.state.update(key, t, e_mult[idx])
                # In debug mode, k_mult is not a single quantity but multiple;
                #   the order of k_mult_fields MUST match the order of the
                #   return signature for rh()
                k_mult_fields = ('f_tsoil', 'f_smsf')
                if all(f in self.DIAGNOSTICS_INDEX for f in k_mult_fields):
                    for k, key in enumerate(self.DIAGNOSTICS_INDEX):
                        if key in k_mult_fields:
                            idx = k_mult_fields.index(key)
                            self.state.update(key, t, k_mult[idx])
                # IMPORTANT: Put e_mult back together again
                e_mult = e_mult[0] * e_mult[1] * e_mult[2] * e_mult[3]
                if 'apar' in self.DIAGNOSTICS_INDEX:
                    self.state.update( # Back-calculating APAR as GPP/(Emult*LUE)
                        'apar', t, np.divide(gpp,
                            np.multiply(e_mult, self.constants.LUE)))

            # Treat Kmult and Emult differently (i.e., break out the former
            #   but not the latter) because that is what L4C Ops does
            if 'e_mult' in self.STATE_INDEX:
                self.state.update('e_mult', t, e_mult)
            if 't_mult' in self.STATE_INDEX:
                self.state.update('t_mult', t, k_mult[0])
            if 'w_mult' in self.STATE_INDEX:
                self.state.update('w_mult', t, k_mult[1])
            # Update the SOC state
            for p in range(1, 4):
                if t == 0:
                    # At time t=0, we have no state to advance, so update
                    #   the state at t=0 based on the initial state
                    delta = np.add(self.state_initial[p-1,0,...], d_soc[p-1])
                    self.state.update(
                        'soc%d' % p, t, delta, bounds = (0, np.inf))
                else:
                    self.state.advance(
                        'soc%d' % p, t, d_soc[p-1], bounds = (0, np.inf))

        if fields_gpp is None:
            fields_gpp = ['apar', 'tmin', 'vpd', 'ft', 'smrz']
        if fields_rh is None:
            fields_rh = ['tsoil', 'smsf']
        with L4CStreamingInputDataset(
                self.file_path, self.CONSTANTS, self.BOUNDS) as hdf:
            num_steps = self.config['time_steps'] if steps is None else steps
            with ProgressBar(num_steps, 'Running...') as progress:
                for t in range(self._time_idx + 1, num_steps):
                    step(t, fields_gpp, fields_rh)
                    self._time_idx += 1
                    progress.update(t)


if __name__ == '__main__':
    # Report on a potential driver dataset with: python main.py data.h5
    import sys
    from pyl4c.data.fixtures import BPLUT

    # Example configuration for all 356 sites in the post-launch period
    config = {
        'bplut': BPLUT,
        'inputs_file_path': sys.argv[1],
        'site_count': sys.argv[2],
        'time_steps': sys.argv[3],
        'start': None
    }

    with h5py.File(config['inputs_file_path'], 'r') as hdf:
        report(hdf, config)
